#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import numpy as np
import subprocess
import signal
import psutil
from ..adaptor import FRAMEWORKS
from ..objective import MultiObjective
from ..conf.config import BenchmarkConf
from ..conf.dotdict import DotDict
from ..utils import logger
from ..utils import OPTIONS
from ..utils.utility import set_backend, GLOBAL_STATE, MODE
from ..utils.create_obj_from_config import create_eval_func, create_dataloader
from ..conf.dotdict import deep_get, deep_set
from ..model import BaseModel
from .data import TRANSFORMS
from .metric import METRICS
from .common import Model as NCModel
from .common import Metric as NCMetric
from .common import Postprocess as NCPostprocess
from .common import _generate_common_dataloader
from ..model.model import get_model_fwk_name

def set_env_var(env_var, value, overwrite_existing=False):
    """Sets the specified environment variable. Only set new env in two cases:
        1. env not exists
        2. env already exists but overwirte_existing params set True
    """
    if overwrite_existing or not os.environ.get(env_var):
        os.environ[env_var] = str(value)

def set_all_env_var(conf, overwrite_existing=False):
    # neural_compressor only use physical cores
    cpu_counts = psutil.cpu_count(logical=False)
    if not conf:
        conf = {}
        conf['num_of_instance'] = 1
        conf['cores_per_instance'] = cpu_counts
    if 'cores_per_instance' in conf:
        assert conf['cores_per_instance'] * conf['num_of_instance'] <= cpu_counts,\
            'num_of_instance * cores_per_instance should <= cpu physical cores'
    else:
        assert conf['num_of_instance'] <= cpu_counts, 'num_of_instance should <= cpu counts'
        conf['cores_per_instance'] = int(cpu_counts / conf['num_of_instance'])
    
    for var, value in conf.items():
        set_env_var(var.upper(), value, overwrite_existing)

class Benchmark(object):
    """Benchmark class can be used to evaluate the model performance, with the objective
       setting, user can get the data of what they configured in yaml
       NOTICE: neural_compressor Benchmark will use the original command to run sub process, this will
       depend on user's code and has possibility to run unneccessary code

    Args:
        conf_fname_or_obj (string or obj): The path to the YAML configuration file or
            BenchmarkConf class containing accuracy goal, tuning objective and preferred
            calibration & quantization tuning space etc.

    """

    def __init__(self, conf_fname_or_obj=None):
        self.framework = None
        self._model = None
        self._b_dataloader = None
        self._b_func = None
        self._metric = None
        self._results = {}
        if isinstance(conf_fname_or_obj, BenchmarkConf):
            self.conf = conf_fname_or_obj
        else:
            self.conf = BenchmarkConf(conf_fname_or_obj)
        if self.conf.usr_cfg.model.framework != 'NA':
            self.framework = self.conf.usr_cfg.model.framework.lower()
            set_backend(self.framework)

    def __call__(self, mode='performance'):
        cfg = self.conf.usr_cfg
        assert cfg.evaluation is not None, 'benchmark evaluation filed should not be None...'
        # use first eval config in yaml if mode from __call__not same with yaml config
        if not mode in cfg.evaluation:
            mode = list(cfg.evaluation.keys())[0]
        assert sys.platform in ['linux', 'win32'], 'only support platform windows and linux...'
        set_all_env_var(deep_get(cfg, 'evaluation.{}.configs'.format(mode)))
        # disable multi-instance for accuracy mode
        if mode == "accuracy":
            set_env_var('NC_ENV_CONF', True, overwrite_existing=True)

        logger.info("Start to run Benchmark.")
        if os.environ.get('NC_ENV_CONF') == 'True':
            return self.run_instance(mode)
        else:
            return self.config_instance()

    fit = __call__

    def config_instance(self):
        raw_cmd = sys.executable + ' ' + ' '.join(sys.argv)
        multi_instance_cmd = ''
        num_of_instance = int(os.environ.get('NUM_OF_INSTANCE'))
        cores_per_instance = int(os.environ.get('CORES_PER_INSTANCE'))
        for i in range(0, num_of_instance):
            core_list = np.arange(0, cores_per_instance) + i * cores_per_instance
            # bind cores only allowed in linux/mac os with numactl enabled
            prefix = self.generate_prefix(core_list)
            instance_cmd = '{} {}'.format(prefix, raw_cmd)
            if sys.platform in ['linux']:
                instance_log = '{}_{}_{}.log'.format(num_of_instance, cores_per_instance, i)
                multi_instance_cmd += '{} 2>&1|tee {} & \\\n'.format(
                    instance_cmd, instance_log)
            else:  # pragma: no cover
                # (TODO) should also add log to win32 benchmark
                multi_instance_cmd += '{} \n'.format(instance_cmd)
        
        multi_instance_cmd += 'wait' if sys.platform in ['linux'] else ''
        logger.info("Running command is\n{}".format(multi_instance_cmd))
        # each instance will execute single instance
        set_env_var('NC_ENV_CONF', True, overwrite_existing=True)
        if sys.platform in ['linux']:
            p = subprocess.Popen(multi_instance_cmd, preexec_fn=os.setsid, shell=True) # nosec
        elif sys.platform in ['win32']:  # pragma: no cover
            p = subprocess.Popen(multi_instance_cmd, start_new_session=True, shell=True) # nosec
        try:
            p.communicate()
        except KeyboardInterrupt:
            os.killpg(os.getpgid(p.pid), signal.SIGKILL)

    def generate_prefix(self, core_list):
        if sys.platform in ['linux'] and os.system('numactl --show >/dev/null 2>&1') == 0:
            return 'OMP_NUM_THREADS={} numactl --localalloc --physcpubind={}'.format(\
                len(core_list), ','.join(core_list.astype(str)))
        elif sys.platform in ['win32']:  # pragma: no cover
            # (TODO) should we move the hw_info from ux?
            from neural_compressor.ux.utils.hw_info import get_number_of_sockets
            num_of_socket = int(get_number_of_sockets())
            cores_per_instance = int(os.environ.get('CORES_PER_INSTANCE'))
            cores_per_socket = int(psutil.cpu_count(logical=False)) / num_of_socket
            socket_id = int(core_list[0] // cores_per_socket)
            # cores per socket should integral multiple of cores per instance, else not bind core
            if cores_per_socket % cores_per_instance == 0:
                from functools import reduce
                hex_core = hex(reduce(lambda x, y : x | y, [1 << p for p in core_list]))
                return 'start /b /WAIT /node {} /affinity {} CMD /c'.format(socket_id, hex_core)
        else:
            return ''

    def run_instance(self, mode):
        cfg = self.conf.usr_cfg
        GLOBAL_STATE.STATE = MODE.BENCHMARK
        framework_specific_info = {'device': cfg.device, \
                                   'approach': cfg.quantization.approach, \
                                   'random_seed': cfg.tuning.random_seed}
        framework = cfg.model.framework.lower()
        if 'tensorflow' in framework:
            framework_specific_info.update({"inputs": cfg.model.inputs, \
                                            "outputs": cfg.model.outputs, \
                                            "recipes": cfg.model.recipes, \
                                            'workspace_path': cfg.tuning.workspace.path})
        if framework == 'mxnet':
            framework_specific_info.update({"b_dataloader": self._b_dataloader})
        if 'onnxrt' in framework.lower():
            framework_specific_info.update(
                                {"backend": framework.lower().split('_')[-1], \
                                 'workspace_path': cfg.tuning.workspace.path, \
                                 'graph_optimization': OPTIONS[framework].graph_optimization})
        if framework == 'pytorch_ipex' or framework == 'pytorch' or framework == 'pytorch_fx':
            framework_specific_info.update({"workspace_path": cfg.tuning.workspace.path,
                                            "q_dataloader": None})

        assert isinstance(self._model, BaseModel), 'need set neural_compressor Model for quantization....'

        adaptor = FRAMEWORKS[framework](framework_specific_info)

        if deep_get(cfg, 'evaluation.{}.iteration'.format(mode)) == -1 and 'dummy_v2' in \
            deep_get(cfg, 'evaluation.{}.dataloader.dataset'.format(mode), {}):
            deep_set(cfg, 'evaluation.{}.iteration'.format(mode), 10)

        iteration = -1 if deep_get(cfg, 'evaluation.{}.iteration'.format(mode)) is None \
            else deep_get(cfg, 'evaluation.{}.iteration'.format(mode))

        metric = [self._metric] if self._metric else \
                    deep_get(cfg, 'evaluation.{}.metric'.format(mode))
        b_postprocess_cfg = deep_get(cfg, 'evaluation.{}.postprocess'.format(mode))

        if self._b_dataloader is None:
            assert deep_get(cfg, 'evaluation.{}.dataloader'.format(mode)) is not None, \
                'dataloader field of yaml file is missing'

            b_dataloader_cfg = deep_get(cfg, 'evaluation.{}.dataloader'.format(mode))
            self._b_dataloader = create_dataloader(self.framework, b_dataloader_cfg)
            
        if self._b_func is None:
            self._b_func = create_eval_func(self.framework, \
                                    self._b_dataloader, \
                                    adaptor, \
                                    metric, \
                                    b_postprocess_cfg,
                                    iteration=iteration)

        objectives = [i.lower() for i in cfg.tuning.multi_objectives.objective] if \
            deep_get(cfg, 'tuning.multi_objectives') else [cfg.tuning.objective]
        assert len(objectives) == 1, 'benchmark supports one objective at a time'
        self.objectives = MultiObjective(objectives,
                              cfg.tuning.accuracy_criterion,
                              is_measure=True)

        val = self.objectives.evaluate(self._b_func, self._model)
        # measurer contain info not only performance(eg, memory, model_size)
        # also measurer have result list among steps
        acc, _ = val
        batch_size = self._b_dataloader.batch_size
        warmup =  0 if deep_get(cfg, 'evaluation.{}.warmup'.format(mode)) is None \
            else deep_get(cfg, 'evaluation.{}.warmup'.format(mode))

        if len(self.objectives.objectives[0].result_list()) < warmup:
            if len(self.objectives.objectives[0].result_list()) > 1 and warmup != 0:
                warmup = 1
            else:
                warmup = 0

        result_list = self.objectives.objectives[0].result_list()[warmup:]
        latency = np.array(result_list).mean() / batch_size
        self._results[mode] = acc, batch_size, result_list

        logger.info("\n{} mode benchmark result:".format(mode))
        for i, res in enumerate(result_list):
            logger.debug("Iteration {} result {}:".format(i, res))
        if mode == 'accuracy':
            logger.info("Batch size = {}".format(batch_size))
            if isinstance(acc, list):
                logger.info("Accuracy is" + "".join([" {:.4f}".format(i) for i in acc]))
            else:
                logger.info("Accuracy is {:.4f}".format(acc))
        elif mode == 'performance':
            logger.info("Batch size = {}".format(batch_size))
            logger.info("Latency: {:.3f} ms".format(latency * 1000))
            logger.info("Throughput: {:.3f} images/sec".format(1. / latency))

    @property
    def results(self):
        return self._results

    @property
    def b_dataloader(self):
        return self._b_dataloader

    @b_dataloader.setter
    def b_dataloader(self, dataloader):
        """Set Data loader for benchmark, It is iterable and the batched data 
           should consists of a tuple like (input, label) or yield (input, _), 
           when b_dataloader is set, user can configure postprocess(optional) and metric 
           in yaml file or set postprocess and metric cls for evaluation.
           Or just get performance without label in dataloader and configure postprocess/metric.

           Args:
               dataloader(generator): user are supported to set a user defined dataloader
                                      which meet the requirements that can yield tuple of
                                      (input, label)/(input, _) batched data.
                                      Another good practice is to use 
                                      neural_compressor.experimental.common.DataLoader
                                      to initialize a neural_compressor dataloader object.
                                      Notice neural_compressor.experimental.common.DataLoader 
                                      is just a wrapper of the information needed to 
                                      build a dataloader, it can't yield
                                      batched data and only in this setter method 
                                      a 'real' eval_dataloader will be created, 
                                      the reason is we have to know the framework info
                                      and only after the Quantization object created then
                                      framework infomation can be known.
                                      Future we will support creating iterable dataloader 
                                      from neural_compressor.experimental.common.DataLoader

        """
        self._b_dataloader = _generate_common_dataloader(dataloader, self.framework)

    @property
    def b_func(self):
        """ not support get b_func """
        assert False, 'Should not try to get the value of `b_func` attribute.'
        return None

    @b_func.setter
    def b_func(self, user_b_func):
        """Eval function for benchmark.

        Args:
            user_b_func: This function takes "model" as input parameter
                         and executes entire training process with self
                         contained training hyper-parameters. If train_func set,
                         an evaluation process must be triggered and user should
                         set eval_dataloader with metric configured or directly eval_func
                         to make evaluation of the model executed.
        """
        self._b_func = user_b_func

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, user_model):
        """Set the user model and dispatch to framework specific internal model object

        Args:
           user_model: user are supported to set model from original framework model format
                       (eg, tensorflow frozen_pb or path to a saved model),
                       but not recommended. Best practice is to set from a initialized
                       neural_compressor.experimental.common.Model.
                       If tensorflow model is used, model's inputs/outputs will be 
                       auto inferenced, but sometimes auto inferenced 
                       inputs/outputs will not meet your requests,
                       set them manually in config yaml file.
                       Another corner case is slim model of tensorflow,
                       be careful of the name of model configured in yaml file,
                       make sure the name is in supported slim model list.
        
        """
        if not isinstance(user_model, BaseModel):
            logger.warning("Force convert framework model to neural_compressor model.")
            self._model = NCModel(user_model)
        else:
            self._model = user_model

        cfg = self.conf.usr_cfg
        if cfg.model.framework == 'NA':
            self.framework = get_model_fwk_name(user_model)
            cfg.model.framework = self.framework
            set_backend(self.framework)  

        # (TODO) ugly to set these params, but tensorflow need
        if 'tensorflow' in self.framework:
            self._model.name = cfg.model.name
            self._model.output_tensor_names = cfg.model.outputs
            self._model.input_tensor_names = cfg.model.inputs
            self._model.workspace_path = cfg.tuning.workspace.path

    @property
    def metric(self):
        assert False, 'Should not try to get the value of `metric` attribute.'
        return None

    @metric.setter
    def metric(self, user_metric):
        """Set metric class and neural_compressor will initialize this class when evaluation
           neural_compressor have many built-in metrics, but user can set specific metric through
           this api. The metric class should take the outputs of the model or 
           postprocess(if have) as inputs, neural_compressor built-in metric always take 
           (predictions, labels) as inputs for update,
           and user_metric.metric_cls should be sub_class of neural_compressor.metric.BaseMetric
           or user defined metric object

        Args:
            user_metric(neural_compressor.experimental.common.Metric):
                user_metric should be object initialized from
                neural_compressor.experimental.common.Metric, in this method the 
                user_metric.metric_cls will be registered to
                specific frameworks and initialized.
                                              
        """
        if deep_get(self.conf.usr_cfg, "evaluation.accuracy.metric"):
            logger.warning("Override the value of `metric` field defined in yaml file" \
                           " as user defines the value of `metric` attribute by code.")
 
        if isinstance(user_metric, NCMetric):
            metric_cfg = {user_metric.name : {**user_metric.kwargs}}
            deep_set(self.conf.usr_cfg, "evaluation.accuracy.metric", metric_cfg)
            self.conf.usr_cfg = DotDict(self.conf.usr_cfg)
            metrics = METRICS(self.framework)
            metrics.register(user_metric.name, user_metric.metric_cls)
        else:
            for i in ['reset', 'update', 'result']:
                assert hasattr(user_metric, i), 'Please realise {} function' \
                                                'in user defined metric'.format(i)
            self._metric = user_metric

    @property
    def postprocess(self, user_postprocess):
        assert False, 'Should not try to get the value of `postprocess` attribute.'
        return None

    @postprocess.setter
    def postprocess(self, user_postprocess):
        """Set postprocess class and neural_compressor will initialize this class when evaluation. 
           The postprocess class should take the outputs of the model as inputs, and
           output (predictions, labels) as inputs for metric update.
           user_postprocess.postprocess_cls should be sub_class of neural_compressor.data.BaseTransform.

        Args:
            user_postprocess(neural_compressor.experimental.common.Postprocess): 
                user_postprocess should be object initialized from
                neural_compressor.experimental.common.Postprocess,
                in this method the user_postprocess.postprocess_cls will be 
                registered to specific frameworks and initialized.

        """
        assert isinstance(user_postprocess, NCPostprocess), \
            'please initialize a neural_compressor.experimental.common.Postprocess and set....'
        postprocess_cfg = {user_postprocess.name : {**user_postprocess.kwargs}}
        if deep_get(self.conf.usr_cfg, "evaluation.accuracy.postprocess"):
            logger.warning("Override the value of `postprocess` field defined in yaml file" \
                           " as user defines the value of `postprocess` attribute by code.")
        deep_set(self.conf.usr_cfg, "evaluation.accuracy.postprocess.transform", postprocess_cfg)
        postprocesses = TRANSFORMS(self.framework, 'postprocess')
        postprocesses.register(user_postprocess.name, user_postprocess.postprocess_cls)

    def __repr__(self):
        return 'Benchmark'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod

class BaseModel:
    ''' base class of all neural_compressor.model, will play graph role'''

    def __init__(self, model, **kwargs): 
        pass

    @property
    def model(self):
        ''' return model itself '''
        raise NotImplementedError

    @property
    def graph_info(self):
        ''' return {Node: Node_type} like {'conv0': 'conv2d'} '''
        raise NotImplementedError
 
    @abstractmethod
    def save(self, root, *args, **kwargs):
        ''' abstract method of model saving'''
        raise NotImplementedError

    @abstractmethod
    def framework(self):
        ''' abstract method of model framework'''
        raise NotImplementedError
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from .pruner import pruner_registry, Pruner
from heapq import heappush, heappop
from ..utils import logger
import re

@pruner_registry
class GradientSensitivityPruner(Pruner):
    def __init__(self, model, local_config, global_config):
        super(GradientSensitivityPruner, self).__init__(model, local_config, global_config)
        self.parameters = local_config.parameters
        self.importance = {}
        self.elementwise_prune = False if local_config.parameters is not None else True

    def on_epoch_begin(self, epoch):
        if epoch == self.start_epoch:
            # register hook for FWK model to get actual input tensor
            self.model.register_forward_pre_hook()
        if self.elementwise_prune:
            self.sparsity = self.update_sparsity(epoch)
            logger.debug("Start pruning in epoch {} with sparsity {}.".
                         format(str(epoch), str(self.sparsity)))
            self.is_last_epoch = epoch == self.end_epoch
            if epoch >= self.start_epoch and epoch <= self.end_epoch:
                self.compute_mask()

    def on_step_begin(self, batch_id):
        if self.elementwise_prune:
            for weight_name in self.weights:
                if weight_name in self.masks:
                    new_weight = self.masks[weight_name].reshape(\
                            np.array(self.model.get_weight(weight_name).shape)) * \
                            np.array(self.model.get_weight(weight_name))
                    self.model.update_weights(weight_name, new_weight)

    def on_epoch_end(self):
        if self.elementwise_prune:
            if self.is_last_epoch:
                for weight_name in self.weights:
                    if weight_name in self.masks:
                        logger.info(
                            "Set {} sparsity with mask {} {} {}.".format(
                                weight_name, str(
                                self.masks[weight_name].size), str(
                                self.masks[weight_name].sum()), str(
                                1 - self.masks[weight_name].sum() /
                                self.masks[weight_name].size)))
                        new_weight = self.masks[weight_name].reshape(\
                                np.array(self.model.get_weight(weight_name).shape)) * \
                                np.array(self.model.get_weight(weight_name))
                        self.model.update_weights(weight_name, new_weight)
        else:
            for weight_name_raw in self.weights:
                for weight_name in self.parse_weight_name(weight_name_raw):
                    self.prune_weight(self.model,
                                      self.importance,
                                      weight_name,
                                      self.parameters)
        if self.is_last_epoch:
            # remove hooks for FWK model to ensure model saving
            self.model.remove_hooks()

    def parse_weight_name(self, weight_name_pattern):
        # check if asterisk is used to match bert layer indexes
        if '*' not in weight_name_pattern:
            yield weight_name_pattern
        else:
            weight_all_names = self.model.get_all_weight_names()
            importance_inputs = self.parameters['importance_inputs']
            for single_weight_name in weight_all_names:
                index_group = re.match(
                        weight_name_pattern.replace('*', '(\d+)'), single_weight_name)
                if index_group is not None:
                    index = index_group.group(1)
                    if self.parameters.get(index) is None:
                        self.parameters['index'] = int(index)
                    # dynamic change importance_inputs with matched index
                    self.parameters['importance_inputs'] = [
                            x.replace('*', index) for x in self.parameters['importance_inputs']]
                    yield single_weight_name
                    # change importance_inputs back
                    self.parameters['importance_inputs'] = importance_inputs

    def on_step_end(self):
        if self.elementwise_prune:
            for weight_name in self.weights:
                self.update_importance_elementwise(self.model,
                                                   self.importance,
                                                   weight_name)
        else:
            for weight_name_raw in self.weights:
                for weight_name in self.parse_weight_name(weight_name_raw):
                    if self.parameters['importance_metric'] == 'abs_gradient':
                        self.update_importance_abs(self.model,
                                                   self.importance,
                                                   weight_name,
                                                   self.parameters)
                    elif self.parameters['importance_metric'] == 'weighted_gradient':
                        self.update_importance_weighted(self.model,
                                                        self.importance,
                                                        weight_name,
                                                        self.parameters)

    def compute_mask(self):
        """compute masks according to absolute values"""
        for weight_name in self.weights:
            if weight_name in self.importance.keys():
                tensor = self.importance[weight_name]
                if len(tensor.shape) in self.tensor_dims:
                    reduced_tensor = self.pattern.reduce(tensor)
                    if self.method == "per_channel":
                        tensor_flat = reduced_tensor.reshape(list(tensor.shape)[:-2], -1)
                        tensor_flat.sort(axis=-1)
                        threshold = tensor_flat[..., int(self.sparsity * tensor_flat.shape[-1])]
                        threshold = np.expand_dims(np.expand_dims(threshold, -1), -1)
                        threshold = np.repeat(threshold, reduced_tensor.shape[-1], axis=-1)
                        threshold = np.repeat(threshold, reduced_tensor.shape[-2], axis=-2)
                    else:
                        tensor_flat = sorted(np.abs(reduced_tensor.flatten()))
                        threshold = float(tensor_flat[int(len(tensor_flat) * self.sparsity)])
                    reduced_mask = threshold < np.abs(reduced_tensor)
                    self.masks[weight_name] = self.pattern.repeat_mask(reduced_mask)

    def prune_weight(self, model, importance, weight_name, parameters):
        if parameters['normalize']:
            exponent = 2
            norm_by_layer = np.power(
                    np.power(importance[weight_name], exponent).sum(-1), 1 / exponent)
            importance[weight_name] /= np.expand_dims(norm_by_layer, -1) + 1e-20
        importance = importance[weight_name]
        weight_tensor = np.array(model.get_weight(weight_name))
        if parameters['transpose']:
            weight_tensor = weight_tensor.transpose((1, 0))

        weight_tensor = self.prune_by_importance(weight_tensor,
                                                 importance,
                                                 parameters['target'],
                                                 parameters['stride'])
        if parameters['transpose']:
            weight_tensor = weight_tensor.transpose((1, 0))

        model.update_weights(weight_name, weight_tensor)

    def update_importance_elementwise(self, model, importance, weight_name):
        if importance.get(weight_name) is not None:
           importance[weight_name] += np.absolute(
                np.array(np.array(model.get_gradient(weight_name)) * np.array(model.get_weight(weight_name))))
        else:
            importance[weight_name] = np.absolute(
                np.array(model.get_gradient(weight_name) * np.array(model.get_weight(weight_name))))

    def update_importance_abs(self, model, importance, weight_name, parameters):
        head_mask = model.get_inputs(
                input_name=parameters['importance_inputs'][0])
        if importance.get(weight_name) is not None:
           importance[weight_name] += np.absolute(
                np.array(model.get_gradient(head_mask)))[parameters['index']]
        else:
            importance[weight_name] = np.absolute(
                np.array(model.get_gradient(head_mask)))[parameters['index']]

    def update_importance_weighted(self, model, importance, weight_name, parameters):
        def weighted_grad(input_weight):
            weight_grad = np.array(model.get_gradient(input_weight))
            weight = np.array(model.get_weight(input_weight))
            weighted_grad = weight_grad * weight
            # TODO: add more check here
            if weighted_grad.ndim > 1:
                weighted_grad = weighted_grad.sum(1)
            return weighted_grad

        accumulated_grad = sum([weighted_grad(input_weight) for input_weight \
                            in parameters['importance_inputs']])

        if importance.get(weight_name) is not None:
            importance[weight_name] += np.absolute(accumulated_grad)
        else:
            importance[weight_name] = np.absolute(accumulated_grad)

    def prune_by_importance(self, tensor, importance, num_instances, stride):
        # structured prune
        importance_ordered = []
        i = 0
        for heads in importance:
            heappush(importance_ordered, (-heads, i))
            i += 1
        sorted_tensor_to_concat = None
        i = 0
        while importance_ordered and i < num_instances:
            head_to_add = heappop(importance_ordered)[1]
            if sorted_tensor_to_concat is None:
                sorted_tensor_to_concat = (
                        tensor[int(head_to_add * stride): int(head_to_add * stride) +
                               int(stride), ...], )
            else:
                sorted_tensor_to_concat += (
                        tensor[int(head_to_add * stride): int(head_to_add * stride) +
                               int(stride), ...], )
            i += 1
        return np.concatenate(sorted_tensor_to_concat)



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# FOR PYTORCH ONLY

from .function import get_all_wrap_children
import pprint
import re
from ..utils.line_operation import get_line_indent_level, of_definition_format
from typing import List
from .. import globals
import logging

logging.basicConfig(level=globals.logging_level,
                    format='%(asctime)s %(levelname)s %(message)s',
                    datefmt='%a, %d %b %Y %H:%M:%S +0000')
logger = logging.getLogger(__name__)


class ClassDefinition:
    # (class_name + file_path) is the unique determination
    def __init__(self, class_name, file_path, class_def_line_idx, parent_class_name):
        self.class_name = class_name
        self.file_path = file_path
        self.class_def_line_idx = class_def_line_idx
        self.parent_class_name = parent_class_name

    def print_info(self):
        logger.debug(f"ClassDefinitionprint_info(): {self.__dict__}")


class ModelDefinition:
    def __init__(self,
                 model_name,
                 class_name,
                 file_path,
                 model_def_line_idx,
                 function_def_line_idx,
                 function_name
                 ):
        self.model_name = model_name
        self.class_name = class_name
        self.file_path = file_path
        self.model_def_line_idx = model_def_line_idx
        self.function_def_line_idx = function_def_line_idx
        self.function_name = function_name

    def print_info(self):
        logger.debug(f"ModelDefinition.print_info(): {self.__dict__}")


# search nnModule classes
def register_nnModule_class():
    logger.info(f"Analyzing nn.Module class definitions in all files ...")
    # search raw nnModule class (e.g. class ClassName(nn.Module):)
    for cl in globals.list_code_line_instance:
        parent_class_has_nnModule = list(set(cl.parent_class_name) & set(
            ["nn.Module", "torch.nn.Module", "nn.Sequential", "torch.Sequential", "_BaseAutoModelClass"])) != []
        if cl.is_class_def_line and parent_class_has_nnModule:
            CD = ClassDefinition(class_name=cl.class_name,
                                 file_path=cl.file_path,
                                 class_def_line_idx=cl.class_def_line_idx,
                                 parent_class_name=cl.parent_class_name)
            CD.print_info()
            globals.list_class_name.append(cl.class_name)
            globals.list_class_def_instance.append(CD)

    # search child class of nnModule class (recursively),  (e.g. class ClassName(ClassNameFather):)
    # this is to complete the nnModule class list
    # e.g. A(nn.Module), B(A), C(B), D(C)
    search_scope = globals.list_class_name
    do_search = True
    while do_search:
        list_child_class_name = []
        for cl in globals.list_code_line_instance:
            parent_class_has_nnModule = list(
                set(cl.parent_class_name) & set(search_scope)) != []
            if cl.is_class_def_line and parent_class_has_nnModule:
                CD = ClassDefinition(class_name=cl.class_name,
                                     file_path=cl.file_path,
                                     class_def_line_idx=cl.class_def_line_idx,
                                     parent_class_name=cl.parent_class_name)
                CD.print_info()
                globals.list_class_name.append(cl.class_name)
                globals.list_class_def_instance.append(CD)
                list_child_class_name.append(cl.class_name)
        search_scope = list_child_class_name
        if len(search_scope) == 0:
            do_search = False

    # unique
    globals.list_class_name = list(set(globals.list_class_name))

    logger.debug(f"class name count: {len(globals.list_class_name)}")
    logger.debug(f"class name list : {globals.list_class_name}")


# search nnModule instance definition
def register_nnModule_instance_definition():
    logger.info(
        f"Analyzing nn.Module instance (model instance) definitions in all files ...")
    # search model definition lines like "model_name = ClassName(xxx)"
    def_cl = []
    for cl in globals.list_code_line_instance:
        if not cl.is_multi_line_comment and not cl.is_single_line_comment_or_empty:
            is_def, lhs, rhs = of_definition_format(cl.line_content)
            if is_def and \
               rhs in globals.list_class_name + ["Module", "Sequential"] and \
               cl.class_name not in globals.list_class_name and \
               "(" not in cl.return_item:
                def_cl.append(cl)

    list_lhs = []
    list_rhs = []
    list_is_in_func = []
    list_func_name = []
    list_return_item = []
    list_file_path = []
    list_line_idx = []
    list_func_def_line_idx = []
    for cl in def_cl:
        is_def, lhs, rhs = of_definition_format(cl.line_content)
        list_lhs.append(lhs)
        list_rhs.append(rhs)
        list_is_in_func.append(cl.is_in_func)
        list_func_name.append(cl.func_name)
        list_return_item.append(cl.return_item)
        list_file_path.append(cl.file_path)
        list_line_idx.append(cl.line_idx)
        list_func_def_line_idx.append(cl.func_def_line_idx)

    # register qualified model's name of lines like "model_name = ClassName(xxx)"
    globals.list_wrapper_base_function_name = []
    for i in range(len(list_lhs)):
        # situation 1: "model = Net()" outside any function
        if not list_is_in_func[i] and "tokenizer" not in list_lhs[i]:
            # register this model
            globals.list_model_name.append(list_lhs[i])
            MD = ModelDefinition(model_name=list_lhs[i],
                                 class_name=list_rhs[i],
                                 file_path=list_file_path[i],
                                 model_def_line_idx=list_line_idx[i],
                                 function_def_line_idx=-1,
                                 function_name="null")  # this MD is for all models defined outside a function
            MD.print_info()
            globals.list_model_def_instance.append(MD)
        elif list_is_in_func[i]:  # situation 2: "model = Net()" is inside a function
            # situation 2-1: the function does not return another model's name, and is not __init__
            if list_return_item[i] not in list_lhs and \
                    list_func_name[i] != "__init__" and "tokenizer" not in list_lhs[i]:
                # register this model
                globals.list_model_name.append(list_lhs[i])
                MD = ModelDefinition(model_name=list_lhs[i],
                                     class_name=list_rhs[i],
                                     file_path=list_file_path[i],
                                     model_def_line_idx=list_line_idx[i],
                                     function_def_line_idx=list_func_def_line_idx[i],
                                     function_name=list_func_name[i])  \
                    # this MD is for all models defined outside a function
                MD.print_info()
                globals.list_model_def_instance.append(MD)
            # situation 2-2: the function returns another model's name
            elif list_return_item[i] in list_lhs:
                globals.list_wrapper_base_function_name.append(
                    list_func_name[i])

    # register function_name like "xxx" in "def xxx() ... return NNModuleClass()"
    for cl in globals.list_code_line_instance:
        if cl.is_in_func and cl.line_idx == cl.func_return_idx \
                and cl.return_item[:cl.return_item.find("(")] in globals.list_class_name:
            globals.list_wrapper_base_function_name.append(cl.func_name)

    # for all base function_name (that returns nnModule instance),
    # find all wrapper function_name of the base wrapper function_name
    globals.list_wrapper_base_function_name = list(
        set(globals.list_wrapper_base_function_name))
    globals.list_wrapper_children_function_name = []
    for i in globals.list_wrapper_base_function_name:
        globals.list_wrapper_children_function_name += get_all_wrap_children(i)
    globals.list_wrapper_all_function_name = globals.list_wrapper_base_function_name + \
        globals.list_wrapper_children_function_name
    globals.list_wrapper_all_function_name = list(
        set(globals.list_wrapper_all_function_name))

    # register function_name like "xxx" in "def xxx() ... model = some_wrapper_function() ... return model"
    for cl in globals.list_code_line_instance:
        if cl.is_in_func and not cl.is_multi_line_comment and not cl.is_single_line_comment_or_empty:
            is_def, lhs, rhs = of_definition_format(cl.line_content)
            if is_def and \
               rhs in globals.list_wrapper_all_function_name and \
               cl.class_name not in globals.list_class_name and \
               cl.return_item == lhs:
                globals.list_wrapper_base_function_name.append(cl.func_name)

    # (again)
    # for all base function_name (that returns nnModule instance),
    # find all wrapper function_name of the base wrapper function_name
    globals.list_wrapper_base_function_name = list(
        set(globals.list_wrapper_base_function_name))
    for i in globals.list_wrapper_base_function_name:
        globals.list_wrapper_children_function_name += get_all_wrap_children(i)
    globals.list_wrapper_all_function_name += globals.list_wrapper_base_function_name + \
        globals.list_wrapper_children_function_name
    globals.list_wrapper_all_function_name = list(
        set(globals.list_wrapper_all_function_name))

    # print all wrapper function names for debug purpose
    logger.debug(
        f"globals.list_wrapper_all_function_name: {globals.list_wrapper_all_function_name}")

    for cl in globals.list_code_line_instance:
        if not cl.is_multi_line_comment and not cl.is_single_line_comment_or_empty and cl.func_name != "__init__":
            is_def, lhs, rhs = of_definition_format(cl.line_content)
            if is_def and \
                rhs in globals.list_wrapper_all_function_name and \
                rhs not in ["self.model", "model", "self.call", "call"] and \
                "forward" not in rhs and \
                "config" not in lhs and "congfig" not in lhs and "," not in lhs and \
                "inference" not in lhs and "tokenizer" not in lhs and \
                cl.class_name not in globals.list_class_name and \
                    cl.func_name not in globals.list_wrapper_all_function_name:
                # register this model
                globals.list_model_name.append(lhs)
                MD = ModelDefinition(model_name=lhs,
                                     class_name="(note: this is a func-defined model)"+rhs,
                                     file_path=cl.file_path,
                                     model_def_line_idx=cl.line_idx,
                                     function_def_line_idx=cl.func_def_line_idx,
                                     function_name=cl.func_name)
                # this MD is for all models defined by a wrapper function (e.g. model = models.resnet18())
                # model def can be outside any function, or in a function that is not itself a wrapper function
                MD.print_info()
                globals.list_model_def_instance.append(MD)

    globals.list_model_name = list(set(globals.list_model_name))

    # print all model names for debug purpose
    logger.debug(f"model name list: {globals.list_model_name}")



#
#  -*- coding: utf-8 -*-
#
#  Copyright (c) 2021 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import os
import numpy as np
from onnx import helper
from onnx import onnx_pb as onnx_proto  
from onnxruntime.quantization.quant_utils import QuantType       

support_pair = {
    'uint8 uint8': True,
    '2 2': True,
    'float16 float16': True,
    '10 10': True,
    'float32 float16': True,
    '1 10': True,
    'float16 float32': True,
    '10 1': True
}

dtype_mapping = {
    'fp32': 1,
    'fp16': 10,
    'int8': 3,
    'uint8': 2,
}

def make_quant_node(name, inputs, outputs):
    return helper.make_node("QuantizeLinear", inputs, outputs, name)

def make_dquant_node(name, inputs, outputs):
    return helper.make_node("DequantizeLinear", inputs, outputs, name)

def _get_qrange_for_qType(qType, reduce_range=False):
    '''
    Helper function to get the quantization range for a type.
        parameter qType: quantization type.
        return: quantization range.
    '''
    if qType == onnx_proto.TensorProto.UINT8:
        return 127 if reduce_range else 255
    elif qType == onnx_proto.TensorProto.INT8:
        # [-64, 64] for reduce_range, and [-127, 127] full_range.
        return 128 if reduce_range else 254
    else:
        raise ValueError('unsupported quantization data type')

def split_shared_bias(model):
    for input_name, node_list in model.input_name_to_nodes.items():
        if len(node_list) > 1 and input_name in [i.name for i in model.model.graph.initializer]:
            for node in node_list[1:]:
                if node.op_type not in ['Conv', 'FusedConv']:
                    continue
                if node.input[2] == input_name:
                    new_input_name = node.input[2] + '_nc_split_' + node.name
                    new_input = helper.make_tensor(
                                    new_input_name,
                                    model.get_initializer(input_name).data_type,
                                    model.get_initializer(input_name).dims,
                                    model.get_initializer(input_name).raw_data,
                                    True)
                    model.add_initializer(new_input)
                    node.input[2] = new_input_name
    return model    

def convert_np_to_float16(np_array, min_positive_val=1e-7, max_finite_val=1e4): # pragma: no cover
    '''
    Convert float32 numpy array to float16 without changing sign or finiteness.
    Positive values less than min_positive_val are mapped to min_positive_val.
    Positive finite values greater than max_finite_val are mapped to max_finite_val.
    Similar for negative values. NaN, 0, inf, and -inf are unchanged.
    '''
    def between(a, b, c):
        return np.logical_and(a < b, b < c)
    np_array = np.where(between(0, np_array, min_positive_val), min_positive_val, np_array)
    np_array = np.where(between(-min_positive_val, np_array, 0), -min_positive_val, np_array)
    np_array = np.where(between(max_finite_val, np_array, float('inf')), max_finite_val, np_array)
    np_array = np.where(between(float('-inf'), np_array, -max_finite_val), -max_finite_val, np_array)
    return np.float16(np_array)

def _npfloat16_to_int(np_list):
    '''
    Convert numpy float16 to python int.
        param np_list: numpy float16 list
        return int_list: python int list
    '''
    return [int(bin(_.view('H'))[2:].zfill(16), 2) for _ in np_list]

def cast_tensor(tensor, dtype, min_positive_val=1e-7, max_finite_val=1e4): # pragma: no cover
    '''
    Convert tensor float to float16.
        param tensor: TensorProto object
        return tensor_float16: converted TensorProto object
    Example:
        from onnxmltools.utils.float16_converter import convert_tensor_float_to_float16
        new_tensor = convert_tensor_float_to_float16(tensor)
    '''
    if not isinstance(tensor, onnx_proto.TensorProto):
        raise ValueError('Expected input type is an ONNX TensorProto but got %s' % type(tensor))

    if tensor.data_type == onnx_proto.TensorProto.FLOAT:
        tensor.data_type = onnx_proto.TensorProto.FLOAT16
        # convert float_data (float type) to float16 and write to int32_data
        if tensor.float_data:
            float16_data = convert_np_to_float16(np.array(tensor.float_data), min_positive_val, max_finite_val)
            int_list = _npfloat16_to_int(float16_data)
            tensor.int32_data[:] = int_list
            tensor.float_data[:] = []
        # convert raw_data (bytes type)
        if tensor.raw_data:
            # convert n.raw_data to float
            float32_list = np.fromstring(tensor.raw_data, dtype='float32')
            # convert float to float16
            float16_list = convert_np_to_float16(float32_list, min_positive_val, max_finite_val)
            # convert float16 to bytes and write back to raw_data
            tensor.raw_data = float16_list.tostring()
    return tensor

def remove_init_from_model_input(model):
    inputs = model.model.graph.input
    name_to_input = {}
    for inp in inputs:
        name_to_input[inp.name] = inp
    for initializer in model.model.graph.initializer:
        if initializer.name in name_to_input:
            inputs.remove(name_to_input[initializer.name])

def collate_preds(results):
    batch = results[0]
    if isinstance(batch, list):
        results = zip(*results)
        collate_results = []
        for output in results:
           collate_results.append(np.concatenate(output))
    elif isinstance(batch, np.ndarray):
        collate_results = np.concatenate(results)
    return collate_results

def quantize_data_with_scale_zero(data, qType, scheme, scale, zero_point):
    '''
        :parameter data: data to quantize
        :parameter qType: data type to quantize to. Supported types UINT8 and INT8
        :parameter scheme: sym or asym quantization.
        :parameter scale: computed scale of quantized data
        :parameter zero_point: computed zero point of quantized data
        :return: quantized weights
        To pack weights, we compute a linear transformation
            - when data type == uint8 mode, from [rmin, rmax] -> [0, 2^{b-1}] and
            - when data type == int8, from [-m , m] -> [-(2^{b-1}-1), 2^{b-1}-1] where
                m = max(abs(rmin), abs(rmax))
    '''
    if qType == onnx_proto.TensorProto.INT8 and scheme == 'sym':
        # signed byte type
        quantized_data = (np.asarray(data) / scale).round().astype('b')
    elif qType == onnx_proto.TensorProto.UINT8 and scheme == 'asym':
        quantized_data = ((np.asarray(data) / scale).round() + zero_point).astype('B')
    else:
        raise ValueError("Unexpected combination of data type {} and scheme {}.".format(
                                                                        qType, scheme))
    return quantized_data

def quantize_data(data, quantize_range, qType, scheme):
    '''
        :parameter data: data to quantize
        :parameter quantize_range: list of data to weight pack.
        :parameter qType: data type to quantize to. Supported types UINT8 and INT8
        :param scheme: sym or asym quantization.
        :return: minimum, maximum, zero point, scale, and quantized weights
        To pack weights, we compute a linear transformation
            - when data type == uint8 mode, from [rmin, rmax] -> [0, 2^{b-1}] and
            - when data type == int8, from [-m , m] -> [-(2^{b-1}-1), 2^{b-1}-1] where
                m = max(abs(rmin), abs(rmax))
        and add necessary intermediate nodes to trasnform quantized weight to full weight
        using the equation r = S(q-z), where
            r: real original value
            q: quantized value
            S: scale
            z: zero point
    '''
    rmin = min(min(data), 0)
    rmax = max(max(data), 0)

    if scheme == 'sym' and qType == onnx_proto.TensorProto.INT8:
        max_range = max(abs(rmin), abs(rmax))
        scale = (float(max_range) * 2) / quantize_range if max_range > 0 else 1
        zero_point = 0
    elif scheme == 'asym' and qType == onnx_proto.TensorProto.UINT8:
        scale = (float(rmax) - rmin) / quantize_range if rmin != rmax else 1
        zero_point = round((0 - rmin) / scale)
    else:
        raise ValueError("Unexpected combination of data type {} and scheme {}.".format(
            qType, scheme))

    quantized_data = quantize_data_with_scale_zero(data, qType, scheme, scale, zero_point)
    return rmin, rmax, zero_point, scale, quantized_data

def quantize_data_per_channel(tensor_value, qType, scheme, scale_value, zo_value):
    channel_count = tensor_value.shape[0] # TBD, default from axis 0
    new_per_channel_tensor_values = []
    for i in range(channel_count):
        per_channel_tensor_value = tensor_value.take(i, 0)
        per_channel_scale_value = scale_value.take(i)
        per_channel_zero_value = zo_value.take(i)
        new_per_channel_tensor_values.append(quantize_data_with_scale_zero(\
                                                       per_channel_tensor_value,
                                                       qType,
                                                       scheme,
                                                       per_channel_scale_value,
                                                       per_channel_zero_value))
    # combine per_channel_data into one
    reshape_dims = list(tensor_value.shape)  # deep copy
    reshape_dims[0] = 1  # only one per channel for reshape
    new_tensor_value = new_per_channel_tensor_values[0].reshape(reshape_dims)
    for i in range(1, channel_count):
        new_per_channel_tensor_value = new_per_channel_tensor_values[i].\
                                                       reshape(reshape_dims)
        new_tensor_value = np.concatenate((new_tensor_value, \
                                           new_per_channel_tensor_value), 0)
    return new_tensor_value

def dequantize_data_with_scale_zero(tensor_value, scale_value, zo_value): # pragma: no cover
    return (tensor_value.astype(np.float32) - zo_value.astype(np.float32)) * scale_value

def dequantize_data(tensor_value, scale_value, zo_value, axis=0): # pragma: no cover
    if scale_value.size == 1:
        return dequantize_data_with_scale_zero(tensor_value, scale_value, zo_value)
    else:
        channel_count = tensor_value.shape[axis] # TBD, default from axis 0
        new_per_channel_tensor_values = []
        for i in range(channel_count):
            per_channel_tensor_value = tensor_value.take(i, 0)
            per_channel_scale_value = scale_value.take(i)
            per_channel_zero_value = zo_value.take(i)
            new_per_channel_tensor_values.append(dequantize_data_with_scale_zero(\
                                                           per_channel_tensor_value,
                                                           per_channel_scale_value,
                                                           per_channel_zero_value))
        # combine per_channel_data into one
        reshape_dims = list(tensor_value.shape)  # deep copy
        reshape_dims[0] = 1  # only one per channel for reshape
        new_tensor_value = new_per_channel_tensor_values[0].reshape(reshape_dims)
        for i in range(1, channel_count):
            new_per_channel_tensor_value = new_per_channel_tensor_values[i].\
                                                           reshape(reshape_dims)
            new_tensor_value = np.concatenate((new_tensor_value, \
                                               new_per_channel_tensor_value), 0)
        return new_tensor_value

class ValueInfo: # pragma: no cover
    def __init__(self,
                 tensor_name,
                 dtype,
                 new_dtype):
        self.tensor_name = tensor_name
        self.dtype = dtype
        self.new_dtype = new_dtype

class QuantizedValue:
    '''
    Represents a linearly quantized value (input\output\intializer)
    '''
    def __init__(self,
                 name,
                 new_quantized_name,
                 scale_name,
                 zero_point_name,
                 quantized_value_type,
                 axis=None,
                 qType=QuantType.QUInt8):
        self.name = name
        self.q_name = new_quantized_name
        self.scale_name = scale_name
        self.zp_name = zero_point_name
        self.value_type = quantized_value_type
        self.axis = axis
        self.qType = qType

class QuantizedInitializer:
    '''
        Represents a linearly quantized weight input from ONNX operators
    '''
    def __init__(self,
                 name,
                 initializer,
                 rmins,
                 rmaxs,
                 zero_points,
                 scales,
                 data=[],
                 quantized_data=[],
                 axis=None,
                 qType=QuantType.QUInt8):
        self.name = name
        self.initializer = initializer  # TensorProto initializer in ONNX graph
        self.rmins = rmins  # List of minimum range for each axis
        self.rmaxs = rmaxs  # List of maximum range for each axis
        # 1D tensor of zero points computed for each axis. scalar if axis is empty
        self.zero_points = zero_points
        self.scales = scales  # 1D tensor of scales computed for each axis. scalar if axis is empty
        self.data = data  # original data from initializer TensorProto
        self.quantized_data = quantized_data  # weight-packed data from data
        # Scalar to specify which dimension in the initializer to weight pack.
        self.axis = axis
        # If empty, single zero point and scales computed from a single rmin and rmax
        self.qType = qType



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .qdq_base_operator import QDQOperatorBase
from .base_operator import QuantOperatorBase
from .base_operator_cast import CastOperatorBase
from neural_compressor.adaptor.ox_utils.util import QuantizedValue

# For operators that support 8bits operations directly, and output could
# reuse input[0]'s type, zeropoint, scale; For example,Transpose, Reshape, etc.

class Direct8BitOp(QuantOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def convert(self):
        node = self.node
        parents = self.quantizer.model.get_parents(node)
        children = self.quantizer.model.get_children(node)
        if len(children) == 0 and len(parents) == 0:
            return

        if any([i.op_type == 'DequantizeLinear' for i in parents]) and \
            any([i.op_type == 'QuantizeLinear' for i in children]):
            for parent in parents:
                if parent.op_type == 'DequantizeLinear':
                    self.node.input[0] = parent.input[0]
                    self.quantizer.remove_nodes.append(parents[0])
                    break
            for child in children:
                if child.op_type == 'QuantizeLinear':
                    self.quantizer.remove_nodes.append(child)
                    self.quantizer.model.replace_input_of_all_nodes(
                        child.output[0], node.output[0] + '_quantized')
            node.output[0] = node.output[0] + '_quantized' 

class QDQDirect8BitOp(QDQOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def quantize(self):
        node = self.node
        if not self.quantizer.is_valid_quantize_weight(node.input[0]):
            return
        self.quantizer.quantize_inputs(self.node, direct_int8=True)
        if not self.disable_qdq_for_node_output or self.quantizer.mode != 'qdq':
            self.quantizer.quantize_outputs(self.node, direct_int8=True)
        node.name = node.name + "_quant"

class DirectCast(CastOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def cast(self):
        node = self.node
        if node.input[0] not in [i.tensor_name for i in self.quantizer.new_value_info.values()]:
            return
        super().cast()
 



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import onnx
from .base_operator import QuantOperatorBase
from .qdq_base_operator import QDQOperatorBase
from onnxruntime.quantization.quant_utils import find_by_name, get_mul_node, \
                                                 QuantizedValueType
from onnx import onnx_pb as onnx_proto
from neural_compressor.adaptor.ox_utils.util import QuantizedValue
'''
    Used when quantize mode is QuantizationMode.IntegerOps.
'''


class MatMulInteger(QuantOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def convert(self):
        node = self.node
        assert (node.op_type == "MatMul")

        parents = self.quantizer.model.get_parents(node)

        inputs = []
        quantized_name = []
        scale = []
        zp = []
        for parent in parents:
            if parent.op_type == 'DequantizeLinear':
                quantized_name.append(parent.input[0])
            else:
                quantized_name.append(parent.output[0])
            if parent.op_type == 'DynamicQuantizeLinear':
                scale.append(parent.output[1])
                zp.append(parent.output[2])
            else:
                scale.append(parent.input[1])
                zp.append(parent.input[2])
        inputs.extend(quantized_name)
        inputs.extend(zp)
        matmul_integer_output = node.output[0] + "_output_quantized"
        matmul_integer_node = onnx.helper.make_node("MatMulInteger",
                                                    inputs,
                                                    [matmul_integer_output], node.name)
        self.quantizer.new_nodes.append(matmul_integer_node)

        # Add cast operation to cast matmulInteger output to float.
        cast_op_output = matmul_integer_output + "_cast_output"
        cast_node = onnx.helper.make_node("Cast", [matmul_integer_output], [cast_op_output],
                                          matmul_integer_output + "_cast",
                                          to=onnx_proto.TensorProto.FLOAT)
        self.quantizer.new_nodes.append(cast_node)

        # Add mul operation to multiply scales of two inputs.
        scales_mul_op = node.name + "_scales_mul"

        scales_mul_node = find_by_name(scales_mul_op, self.quantizer.new_nodes)
        if scales_mul_node is None:
            scales_mul_node = get_mul_node([scale[0], scale[1]], 
                scales_mul_op + ":0", scales_mul_op)
            self.quantizer.new_nodes.append(scales_mul_node)

        scales_mul_op_output = scales_mul_node.output[0]

        # Add mul operation to multiply mul_scales_op result with output of MatMulInteger
        # and make the output of this node the same as output of original matmul node.
        output_scale_mul_op = node.name + "_output_scale_mul"
        self.quantizer.new_nodes.append(get_mul_node([cast_op_output, scales_mul_op_output], 
                                  node.output[0], 
                                  output_scale_mul_op))
        if parents[1].op_type == 'DequantizeLinear':
            self.quantizer.remove_nodes.append(parents[1])
        self.quantizer.remove_nodes.append(node)

'''
    Used when quantize mode is QuantizationMode.QLinearOps
'''


class QLinearMatMul(QuantOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def convert(self):
        node = self.node
        assert (node.op_type == "MatMul")

        parents = self.quantizer.model.get_parents(node)
        if len(self.quantizer.model.get_children(node)) == 0:
            return
        child = self.quantizer.model.get_children(node)[0]

        qlinear_matmul_output = child.output[0]

        qlinear_matmul_inputs = []
        for parent in parents:
            qlinear_matmul_inputs.extend(parent.input)
        qlinear_matmul_inputs.extend(child.input[1:])

        qlinear_matmul_node = onnx.helper.make_node("QLinearMatMul", 
                                                    qlinear_matmul_inputs, 
                                                    [qlinear_matmul_output],
                                                    node.name)
        self.quantizer.new_nodes.append(qlinear_matmul_node)
        self.quantizer.remove_nodes.extend(parents)
        self.quantizer.remove_nodes.append(child)
        self.quantizer.remove_nodes.append(node)

class QDQMatMul(QDQOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def quantize(self):
        node = self.node
        assert (node.op_type == "MatMul")

        self.quantizer.quantize_inputs(node, [0])
        if self.per_channel and find_by_name(node.input[1], self.quantizer.model.initializer()):
            self.quantizer.quantize_weights_per_channel(node, [1],
                                    self.weight_dtype, self.weight_scheme, -1)
        else:
            self.quantizer.quantize_inputs(node, [1])

        if not self.disable_qdq_for_node_output or self.quantizer.mode != 'qdq':
            self.quantizer.quantize_outputs(node)
        node.name = node.name + "_quant"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from itertools import permutations

from ..conf.config import Conf
from ..utils import logger
from .common import Model as NCModel
from ..model import BaseModel
from .common import Metric, Postprocess
from ..strategy import STRATEGIES

from .quantization import Quantization
from .pruning import Pruning
from .distillation import Distillation
from .model_conversion import ModelConversion
from .graph_optimization import Graph_Optimization
from ..utils.create_obj_from_config import create_dataloader, create_train_func, create_eval_func
from .benchmark import Benchmark
from .component import Component
from ..conf.dotdict import DotDict, deep_set, deep_get

SUPPORTED_COMPONENTS = [
  Quantization,
  Pruning,
  Graph_Optimization,
  ModelConversion,
  Benchmark,
  Component
]

class Scheduler(object):
    """Scheduler for neural_compressor component pipeline execution.

    Neural Compressor supports serveral seperate components: Quantization, Pruning, Benchmarking.
    This scheduler will sequentially execute specified components by the order of
    appending. This interface provids an unique entry to pipeline execute all supported
    components.

    There are two typical usages:
    1) if all informations are set in user configuration yaml files by using neural_compressor built-in
       dataloaders/datasets/metrics, the code usage is like below:

       prune = Pruning('/path/to/pruning.yaml')
       quantizer = Quantization('/path/to/quantization.yaml')
       scheduler = Scheduler()
       scheduler.model('/path/to/model')
       scheduler.append(prune)
       scheduler.append(quantizer)
       opt_model = scheduler()
       opt_model.save()


    2) if neural_compressor built-in dataloaders/datasets/metrics could not fully meet user requirements,
       customized dataloaders/datasets/metrics are needed, the code usage is like below:

       prune = Pruning('/path/to/pruning.yaml')
       prune.train_func = ...                   # optional if it is configured in user yaml.
       prune.eval_dataloader = ...                # optional if it is configured in user yaml.
       prune.eval_func = ...                      # optional if it is configured in user yaml.

       quantizer = Quantization('/path/to/quantization.yaml')
       quantizer.metric = ...                     # optional if it is configured in user yaml.
       quantizer.calib_dataloader = ...           # optional if it is configured in user yaml.
       quantizer.eval_dataloader = ...            # optional if it is configured in user yaml.

       scheduler = Scheduler()
       scheduler.model('/path/to/model')
       scheduler.append(prune)
       scheduler.append(quantizer)
       opt_model = scheduler()
       opt_model.save()

    """

    def __init__(self):
        self._model = None
        self._train_func = None
        self._eval_func = None
        self.components = []

    def append(self, *args):
        """Add neural_compressor component into pipeline for sequential execution.

           Args:
               conf_fname_or_obj (string or obj): The path to user configuration yaml file or
                                                  conf class.
               calib_dataloader (generator): Optional. Data loader for calibration of Post-
                                             Training Static Quantization,
                                             or None for Post-Training Dynamic Quantization,
                                             or Data loader for training phase of Quantization-
                                             aware Training,
                                             or Data loader for training phase of Pruning,
                                             if its corresponding field is not configured in yaml.
               eval_dataloader (generator):  Optional. Data loader for evaluation phase of all
                                             neural_compressor components, if its corresponding field is not
                                             configured in yaml and eval_func is not specified.
               postprocess (Postprocess):    Optional. Object initialized from common.Postprocess.
               metric (Metric):              Optional. Object initialized from common.Metric.
               q_func (func):                Optional. Training function for Quantization-Aware
                                             Training and Pruning cases if user doesn't provide
                                             training info in user configuration yaml file.
               eval_func (func):             Optional. Evaluation function for Quantization-Aware
                                             Training and Pruning, being None for other cases.
               kwargs (named arguments):     Reserved for interface extension.
        """
        for item in args:
            assert any([isinstance(item, supported_component) \
                        for supported_component in SUPPORTED_COMPONENTS])
            self.components.append(item)

    def __call__(self):
        """The main entry point of sequential pipeline execution.

        NOTE: it's user responsibility to ensure the output of each componenet could be fed as
        the input of next component.

        Returns:
            optimized model: best optimized model generated, otherwise return None

        """
        assert self.model, "Scheduler class's model property should be set " \
                           "before invoking this __call__() function"
        model = self.model
        assert len(self.components) > 0
        logger.info("Start sequential pipeline execution.")
        for i, component in enumerate(self.components):
            # print appropriate ordinal number representation (1st, 2nd, 3rd) for each step
            ordinal = lambda n: "%d%s" % (n,"tsnrhtdd"[(n//10%10!=1)*(n%10<4)*n%10::4])
            logger.info("The {} step being executing is {}.".format(
                        ordinal(i),
                        repr(component).upper()
                        ))

            component.model = model
            if self._train_func is not None:
                component.train_func = self._train_func
            if self._eval_func is not None:
                component.eval_func = self._eval_func
            model = component()

        return model

    fit = __call__

    def combine(self, *args):
        """Combine neural_compressor components into a new component.
           Args:
               args (Component): Components to be combined together. Input Component should be
               supported in Neural Compressor and pass the sanity check during combine. The illegal combination
               (e.g. Components uses different frameworks) returns an error.

            Returns:
                Combined Component: The return component is created as base Component class.
                It syncs input components configuration and creates dataloaders/functions
                accordingly.
        """
        assert len(args) > 1, "Combine requires at least 2 components. Please check your inputs."
        # check if input components are good for combine
        self._combination_sanity_check(*args)
        # create component for the combination
        combination = []
        for arg in args:
            combination += [arg.__class__.__name__] \
                    if arg.combination is None else arg.combination
        new_component = Component(combination=combination)
        self._combine_components(*args, dist_component=new_component)

        return new_component

    def _combination_sanity_check(self, *args):
        TEMP_SUPPORTED_COMPONENTS = ['Quantization', 'Pruning', 'Distillation']
        checked_components = []
        for component in args:
            component_class = component.__class__.__name__
            if component_class in TEMP_SUPPORTED_COMPONENTS and \
                    component_class not in checked_components :
                checked_components.append(component_class)
            else:
                logger.error("The combination of {} is not supported.".format(
                    checked_components + [component_class]))

    def _combine_components(self, *args, dist_component=None):
        """Actual implementation of combine(). It loops input component and sync-up status
           to distance component.

           Args:
               args(tuple): input components.
               dist_component(Component): the distance component for the combination

           Returns: None
        """
        assert len(args) > 0, "Empty input detected in combine."
        framework = args[0].framework
        device = args[0].cfg.device
        train_cfg = DotDict()
        eval_cfg = DotDict()
        tuning_cfg = DotDict()
        model_cfg = DotDict()
        quantization_cfg = DotDict()
        has_distillation = False
        for combine_component in args:
            if isinstance(combine_component, Distillation):
                has_distillation = True

        for combine_component in args:
            # check if config is valid
            assert combine_component.framework == framework, "Combined components should have " \
                    "same framework. Detect different frameworks: {} and {} are used.".format(
                    framework, combine_component.framework )
            assert combine_component.cfg.device == device, "Combined components should have " \
                    "same device. Detect different device: {} and {} are used.".format(
                    device, combine_component.cfg.device )

            # sync configs
            component_name = combine_component.__class__.__name__.lower()
            component_cfg = getattr(combine_component.cfg,
                                    component_name,
                                    None)
            assert combine_component is not None, "Please ensure field {} is configured " \
                    "in input yaml".format(component_name)

            # in case of key train/evaluation not exist, return an empty DotDict
            component_train_cfg = component_cfg.get('train', DotDict())

            # TODO: Assumption here: train phase is defined inside component yaml field.
            # But eval is defined at root yaml field.
            component_eval_cfg = combine_component.cfg.get('evaluation', DotDict())
            component_tuning_cfg = combine_component.cfg.get('tuning', DotDict())
            component_model_cfg = combine_component.cfg.get('model', DotDict())
            component_quantization_cfg = combine_component.cfg.get('quantization', DotDict()) \
                if component_name == 'quantization' else DotDict()

            combine_component._model = self._model
            if component_eval_cfg and component_train_cfg:
                # create attributes if eval and train are defined in component config
                combine_component.pre_process()
            elif isinstance(combine_component, Pruning):
                # if no eval/train configuration. Make minimal initialization
                combine_component.generate_hooks()
                combine_component.generate_pruners()
            elif isinstance(combine_component, Distillation):
                # if no eval/train configuration. Make minimal initialization
                combine_component.create_criterion()
                combine_component.create_optimizer()
                combine_component.generate_hooks()

            # If distillation is combined, set training configs
            if has_distillation:
                if isinstance(combine_component, Distillation):
                    component_train_cfg.criterion = combine_component.train_cfg.criterion
                    component_train_cfg.optimizer = combine_component.train_cfg.optimizer
                    dist_component.criterion = combine_component.train_cfg.criterion
                    dist_component.on_after_compute_loss = combine_component.on_after_compute_loss
                    dist_component.teacher_model = combine_component.teacher_model
                    self._sync_config(train_cfg, component_train_cfg)
                # Only sync train_cfg of distillation because criterion should include only one element
                else:
                    component_train_cfg = DotDict()

            self._sync_config(train_cfg, component_train_cfg)
            self._sync_config(eval_cfg, component_eval_cfg)
            self._sync_config(tuning_cfg, component_tuning_cfg)
            self._sync_config(model_cfg, component_model_cfg)
            self._sync_config(quantization_cfg, component_quantization_cfg)

            # sync hooks
            if combine_component.hooks is None:
                continue
            for k, v in combine_component.hooks.items():
                dist_component.register_hook(k, v)

        # sync to dist component
        dist_component_cfg = DotDict()
        if dist_component is not None:
            deep_set(dist_component_cfg, 'train', train_cfg)
            deep_set(dist_component_cfg, 'evaluation', eval_cfg)
            deep_set(dist_component_cfg, 'tuning', tuning_cfg)
            deep_set(dist_component_cfg, 'device', device)
            deep_set(dist_component_cfg, 'model', model_cfg)
            deep_set(dist_component_cfg, 'quantization', quantization_cfg)
            dist_component._model = self._model
            dist_component.framework = framework
            dist_component.cfg = dist_component_cfg

    def _sync_config(self, dist_config, src_config):
        """Sync the configuration between src and dist. It updates the missing keys in dist config
           from src config. And check the value for same keys between src and dist.
           Args:
               dist_config(DotDict): dist configuration to sync
               src_config(DotDict): src configuration to sync

            Returns: None
        """
        # sync the config if dist and src configs are not empty
        if dist_config and src_config:
            # update missing keys in dist_config
            for key in src_config.keys() - dist_config.keys():
                dist_config[key] = src_config[key]
            # check the same keys in dist and src whether is valid
            for key in src_config.keys() & dist_config.keys():
                if isinstance(dist_config[key], dict) and isinstance(src_config[key], dict):
                    self._sync_config(dist_config[key], src_config[key])
                elif dist_config[key] != src_config[key]:
                    logger.warning("Find different value {} and {} on key {}.".format(
                                   dist_config[key], src_config[key], key) + \
                                   " Use first key-value ({}: {}) pair as default".format(
                                   key, dist_config[key]))
        # update src config to dist if dist is empty.
        elif not dist_config and src_config:
            dist_config.update(src_config)

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, user_model):
        """Set the user model and dispatch to framework specific internal model object

        Args:
           user_model: user are supported to set model from original framework model format
                       (eg, tensorflow frozen_pb or path to a saved model),
                       but not recommended. Best practice is to set from a initialized
                       neural_compressor.experimental.common.Model.
                       If tensorflow model is used, model's inputs/outputs will be
                       auto inferenced, but sometimes auto inferenced
                       inputs/outputs will not meet your requests,
                       set them manually in config yaml file.
                       Another corner case is slim model of tensorflow,
                       be careful of the name of model configured in yaml file,
                       make sure the name is in supported slim model list.

        """
        if not isinstance(user_model, BaseModel):
            logger.warning("Force convert framework model to neural_compressor model.")
            self._model = NCModel(user_model)
        else:
            self._model = user_model

    @property
    def train_func(self):
        """ not support get train_func """
        assert False, 'Should not try to get the value of `train_func` attribute.'
        return None

    @train_func.setter
    def train_func(self, user_train_func):
        """Training function.

        Args:
            user_training_func: This function takes "model" as input parameter
                         and executes entire training process with self
                         contained training hyper-parameters. If training_func set,
                         an evaluation process must be triggered and user should
                         set eval_dataloader with metric configured or directly eval_func
                         to make evaluation of the model executed. training_func will return
                         a trained model.
        """
        self._train_func = user_train_func

    @property
    def eval_func(self):
        """ not support get eval_func """
        assert False, 'Should not try to get the value of `eval_func` attribute.'
        return None

    @eval_func.setter
    def eval_func(self, user_eval_func):
        """evaluation function.

        Args:
            user_eval_func: This function takes "model" as input parameter
                         and executes entire evaluation process with self
                         contained metrics. If eval_func set,
                         an evaluation process must be triggered 
                         to make evaluation of the model executed.
        """
        self._eval_func = user_eval_func



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf
from tensorflow.python.framework import tensor_util
from tensorflow.core.framework import attr_value_pb2
from tensorflow.core.framework import node_def_pb2
from tensorflow.python.framework import dtypes

from ..graph_base import GraphRewriterBase
from neural_compressor.adaptor.tf_utils.graph_util import GraphAnalyzer
from neural_compressor.adaptor.tf_utils.graph_util import GraphRewriterHelper as Helper

class FuseConvRequantizeTransformer(GraphRewriterBase):
    """Fuse Quantized Conv Op with the successor Requantize Op.
    """
    fuse_patterns = [[
        "QuantizedConv2DWithBiasAndRelu",
        "QuantizedDepthwiseConv2DWithBiasAndRelu",
        "QuantizedConv2DWithBias",
        "QuantizedDepthwiseConv2DWithBias",
        "_QuantizedConv2D",
        "_QuantizedDepthwiseConv2D",
        "_QuantizedConv3D",
    ], ['RequantizePerChannel', 'Requantize'], ('Dequantize',)]
   
    fuse_sum_op_types = (
        [b'BiasAdd', b'Sum'],
        [b'BiasAdd', b'Sum', b'Relu'],
        [b'BiasAdd', b'Relu', b'Sum'],
        [b'BiasAdd', b'LeakyRelu', b'Sum']
        )

    sum_pattern = [["QuantizedConv2DWithBiasSumAndRelu", "QuantizedConv2DWithBiasReluAndSum", \
                    "_QuantizedDepthwiseConv2D", "_QuantizedConv2D", "_QuantizedConv3D"],
                     ['RequantizePerChannel', 'Requantize']]

    def __init__(self, model, device='cpu', new_api=False):
        super().__init__(model)
        self.device = device
        self.graph_analyzer = GraphAnalyzer()
        self.graph_analyzer.graph = self.model
        self.graph_info = self.graph_analyzer.parse_graph()
        self.fused_ops = []
        self.output_types = []
        self.new_api = new_api

    def do_transformation(self):
        """Fuse the quantized op with the following requantize op.
            The transformation has two stages, the first step is to fuse the patterns
            defined in self.fuse_patterns and the last step is to fuse the self.sum_patterns.
        Returns:
            [graphdef]: the optimized graphdef object
        """
        int8_type = dtypes.qint8.as_datatype_enum
        uint8_type = dtypes.quint8.as_datatype_enum
        float32_type = dtypes.float32.as_datatype_enum
        qint32_type = dtypes.qint32.as_datatype_enum
        dtype_map_dict = {
            dtypes.qint8.as_datatype_enum: dtypes.qint8,
            dtypes.quint8.as_datatype_enum: dtypes.quint8,
            dtypes.float32.as_datatype_enum: dtypes.float32,
            dtypes.qint32.as_datatype_enum: dtypes.qint32
        }
        target_nodes = self.graph_analyzer.query_fusion_pattern_nodes(self.fuse_patterns)

        for i in target_nodes:
            quantized_node_name = i[0]
            quantized_node = self.graph_info[quantized_node_name].node
            if not self.new_api and quantized_node.op == "QuantizedDepthwiseConv2DWithBias":
                continue
            if i[-1][0] in ('_QuantizedDepthwiseConv2D', '_QuantizedConv2D', '_QuantizedConv3D'):
                if str(quantized_node.attr['fused_ops'].list.s).find('Sum') != -1:
                    continue
                #else:
                #   print(quantized_node.attr['fused_ops'].list.s)
            requantize_node_name = i[1]
            requantize_node = self.graph_info[requantize_node_name].node
            if i[-1][-1] == 'Dequantize' and self.new_api and \
                i[0] in ('_QuantizedDepthwiseConv2D', '_QuantizedConv2D', '_QuantizedConv3D'):
                dequantize_node_name = i[2]
            else:
                dequantize_node_name = None
            requested_output_min_name = requantize_node.input[3]
            requested_output_max_name = requantize_node.input[4]

            quantized_node_op = i[-1][0]

            new_node = node_def_pb2.NodeDef()
            if self.new_api:
                if i[-1][0] == 'QuantizedConv2DWithBiasAndRelu':
                    new_node.op = '_QuantizedConv2D'
                    self.fused_ops= [b"BiasAdd", b"Relu", b"Requantize"]
                    self.output_types= [
                                        requantize_node.attr['out_type'].type,
                                        dtypes.float32.as_datatype_enum,
                                        dtypes.float32.as_datatype_enum]
                elif i[-1][0] == 'QuantizedConv2DWithBias':
                    new_node.op = '_QuantizedConv2D'
                    self.fused_ops = [b"BiasAdd", b"Requantize"]
                    self.output_types = [
                                        requantize_node.attr['out_type'].type,
                                        dtypes.float32.as_datatype_enum,
                                        dtypes.float32.as_datatype_enum]
                elif i[-1][0] == 'QuantizedDepthwiseConv2DWithBias':
                    new_node.op = '_QuantizedDepthwiseConv2D'
                    self.fused_ops = [b"BiasAdd", b"Requantize"]
                    self.output_types = [
                                        requantize_node.attr['out_type'].type,
                                        dtypes.float32.as_datatype_enum,
                                        dtypes.float32.as_datatype_enum]
                elif i[-1][0] == 'QuantizedDepthwiseConv2DWithBiasAndRelu':
                    new_node.op = '_QuantizedDepthwiseConv2D'
                    self.fused_ops= [b"BiasAdd", b"Relu", b"Requantize"]
                    self.output_types= [
                                        requantize_node.attr['out_type'].type,
                                        dtypes.float32.as_datatype_enum,
                                        dtypes.float32.as_datatype_enum]
                elif quantized_node_op == '_QuantizedConv2D':
                    new_node.op = '_QuantizedConv2D'
                elif quantized_node_op == '_QuantizedDepthwiseConv2D':
                    new_node.op = '_QuantizedDepthwiseConv2D'
                elif quantized_node_op == '_QuantizedConv3D':
                    new_node.op = '_QuantizedConv3D'
            else:
                new_node.op = quantized_node_op + "AndRequantize"
            new_node.name = requantize_node_name
            for _, value in enumerate(quantized_node.input):
                new_node.input.append(value)

            new_node.input.append(requested_output_min_name)
            new_node.input.append(requested_output_max_name)
            if 'Tinput' in quantized_node.attr:
                new_node.attr["Tinput"].CopyFrom(quantized_node.attr['Tinput'])
            if 'Tfilter' in quantized_node.attr:
                new_node.attr["Tfilter"].CopyFrom(quantized_node.attr['Tfilter'])
            if 'strides' in quantized_node.attr:
                new_node.attr["strides"].CopyFrom(quantized_node.attr['strides'])
            if 'padding' in quantized_node.attr:
                new_node.attr["padding"].CopyFrom(quantized_node.attr['padding'])
            if 'alpha' in quantized_node.attr:
                new_node.attr["alpha"].CopyFrom(quantized_node.attr['alpha'])
            if 'Tsummand' in quantized_node.attr:
                new_node.attr["Tsummand"].CopyFrom(quantized_node.attr['Tsummand'])

            parent_node_name = Helper.node_name_from_input(quantized_node.input[0])
            max_filter_node = self.graph_info[new_node.input[6]].node
            min_filter_node = self.graph_info[new_node.input[5]].node
            last_node = self.graph_info[new_node.input[0]].node
            
            if last_node.op.find('Requantize') != -1 or ((last_node.op.find('QuantizeV2') != -1 or \
                                                            last_node.op.find('QuantizedConv2D') != -1) \
                                                                and len(quantized_node.attr['fused_ops'].list.s) > 0):          
                bias_node = self.graph_info[new_node.input[2]].node
                max_input_node = self.graph_info[last_node.input[-1]].node
                min_input_node = self.graph_info[last_node.input[-2]].node
                min_input = (min_input_node.attr['value'].tensor.float_val)[0]
                max_input = (max_input_node.attr['value'].tensor.float_val)[0]
                if 'Depthwise' in quantized_node_op or requantize_node.op.find('PerChannel') != -1:
                    channel_size = max_filter_node.attr['value'].tensor.tensor_shape.dim[0].size
                    max_filter_tensor = tensor_util.MakeNdarray(
                        min_filter_node.attr['value'].tensor)
                    min_filter_tensor = tensor_util.MakeNdarray(
                        min_filter_node.attr['value'].tensor)
                else:
                    channel_size = 1
                    max_filter_tensor = []
                    min_filter_tensor = []
                    max_filter_tensor.append((max_filter_node.attr['value'].tensor.float_val)[0])
                    min_filter_tensor.append((min_filter_node.attr['value'].tensor.float_val)[0])
                bias_tensor = tensor_util.MakeNdarray(
                    self.graph_info[new_node.input[2]].node.attr['value'].tensor)

                activation_range = 127.0 if new_node.attr["Tinput"].type == dtypes.qint8 else 255.0

                int32_bias = Helper.generate_int32_bias_for_conv(
                    bias_tensor, channel_size, max_input, min_input,
                    max_filter_tensor, min_filter_tensor, activation_range)

                bias_node.attr['dtype'].CopyFrom(
                    attr_value_pb2.AttrValue(
                        type=float32_type if self.device == 'gpu' else qint32_type))

                bias_node.attr['value'].CopyFrom(
                    attr_value_pb2.AttrValue(tensor=tensor_util.make_tensor_proto(
                        bias_tensor if self.device == 'gpu' else int32_bias, dtypes.
                        float32 if self.device == 'gpu' else dtypes.int32, bias_tensor.shape)))

                bias_node.attr['value'].tensor.dtype = float32_type \
                                        if self.device == 'gpu' else qint32_type
                new_node.attr["Tbias"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type \
                                                if self.device == 'gpu' else qint32_type))
            else:
                new_node.attr["Tbias"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type))
            # in tf 2.10, the "padding_list" attr name changes to explicit_paddings
            if "padding_list" in quantized_node.attr:
                if not self.new_api:
                    new_node.attr["padding_list"].CopyFrom(quantized_node.attr['padding_list'])
                elif quantized_node.attr["padding"].s == b"EXPLICIT":
                    new_node.attr["explicit_paddings"].CopyFrom(quantized_node.attr['padding_list'])
            elif "explicit_paddings" in quantized_node.attr:
                new_node.attr["explicit_paddings"].CopyFrom(quantized_node.attr['explicit_paddings'])           
            if "dilations" in quantized_node.attr:
                new_node.attr["dilations"].CopyFrom(quantized_node.attr['dilations'])
            
            if self.new_api and new_node.op in ('_QuantizedConv2D', '_QuantizedDepthwiseConv2D'):
                input_data_type = dtypes.qint8 if new_node.attr["Tinput"].type == dtypes.qint8 else dtypes.quint8
                Helper.set_attr_type_list(new_node, 'input_types', [
                    input_data_type.as_datatype_enum,
                    dtypes.qint8.as_datatype_enum,
                    dtypes.float32.as_datatype_enum if new_node.attr["Tbias"].type == dtypes.float32 \
                                                                            else dtypes.qint32.as_datatype_enum,
                    dtypes.float32.as_datatype_enum,
                    dtypes.float32.as_datatype_enum,
                    dtypes.float32.as_datatype_enum,
                    dtypes.float32.as_datatype_enum,
                    dtypes.float32.as_datatype_enum,
                    dtypes.float32.as_datatype_enum,
                ])

                if quantized_node_op not in ('_QuantizedConv2D', '_QuantizedDepthwiseConv2D'): 
                    Helper.set_attr_type_list(new_node, 'out_types', self.output_types)
                    new_node.attr["Tsummand"].CopyFrom(attr_value_pb2.AttrValue(type=self.output_types[0]))
                else:
                    if str(quantized_node.attr['fused_ops'].list.s) == str([b"BiasAdd", b"_FusedHardSwish"]):
                        self.fused_ops= [b"BiasAdd", b"_FusedHardSwish", b"Requantize"]
                        Helper.set_attr_type_list(new_node, 'out_types', [
                                              requantize_node.attr['out_type'].type,
                                              dtypes.float32.as_datatype_enum,
                                              dtypes.float32.as_datatype_enum ])
                        Helper.set_attr_dtype(new_node, "out_type", \
                            dtype_map_dict[requantize_node.attr['out_type'].type])
                        Helper.set_attr_dtype(new_node, "Tsummand", \
                            dtype_map_dict[requantize_node.attr['out_type'].type])
                    elif str(quantized_node.attr['fused_ops'].list.s) == str([b"BiasAdd", b"_FusedSwish"]):
                        self.fused_ops= [b"BiasAdd", b"_FusedSwish", b"Requantize"]
                        Helper.set_attr_type_list(new_node, 'out_types', [
                                              requantize_node.attr['out_type'].type,
                                              dtypes.float32.as_datatype_enum,
                                              dtypes.float32.as_datatype_enum ])
                        Helper.set_attr_dtype(new_node, "out_type", \
                            dtype_map_dict[requantize_node.attr['out_type'].type])
                        Helper.set_attr_dtype(new_node, "Tsummand", \
                            dtype_map_dict[requantize_node.attr['out_type'].type])
                    elif str(quantized_node.attr['fused_ops'].list.s) == str([b"BiasAdd", b"Relu"]):
                        self.fused_ops= [b"BiasAdd", b"Relu", b"Requantize"]
                        Helper.set_attr_type_list(new_node, 'out_types', [
                                              requantize_node.attr['out_type'].type,
                                              dtypes.float32.as_datatype_enum,
                                              dtypes.float32.as_datatype_enum ])
                        Helper.set_attr_dtype(new_node, "out_type", \
                            dtype_map_dict[requantize_node.attr['out_type'].type])
                        Helper.set_attr_dtype(new_node, "Tsummand", \
                            dtype_map_dict[requantize_node.attr['out_type'].type])
                    elif str(quantized_node.attr['fused_ops'].list.s) == str([b"BiasAdd", b"LeakyRelu"]):
                        self.fused_ops= [b"BiasAdd", b"LeakyRelu", b"Requantize"]
                        Helper.set_attr_type_list(new_node, 'out_types', [
                                              requantize_node.attr['out_type'].type,
                                              dtypes.float32.as_datatype_enum,
                                              dtypes.float32.as_datatype_enum ])
                        Helper.set_attr_dtype(new_node, "out_type", \
                            dtype_map_dict[requantize_node.attr['out_type'].type])
                        Helper.set_attr_dtype(new_node, "Tsummand", \
                            dtype_map_dict[requantize_node.attr['out_type'].type])
                    elif str(quantized_node.attr['fused_ops'].list.s) == str([b"BiasAdd", b"Elu"]):
                        self.fused_ops= [b"BiasAdd", b"Elu", b"Requantize"]
                        Helper.set_attr_type_list(new_node, 'out_types', [
                                              requantize_node.attr['out_type'].type,
                                              dtypes.float32.as_datatype_enum,
                                              dtypes.float32.as_datatype_enum ])
                        Helper.set_attr_dtype(new_node, "out_type", \
                            dtype_map_dict[requantize_node.attr['out_type'].type])
                        Helper.set_attr_dtype(new_node, "Tsummand", \
                            dtype_map_dict[requantize_node.attr['out_type'].type])
                    elif str(quantized_node.attr['fused_ops'].list.s) == str([b"BiasAdd"]):
                        self.fused_ops= [b"BiasAdd", b"Requantize"]
                        Helper.set_attr_type_list(new_node, 'out_types', [
                                              requantize_node.attr['out_type'].type,
                                              dtypes.float32.as_datatype_enum,
                                              dtypes.float32.as_datatype_enum ])
                        Helper.set_attr_dtype(new_node, "out_type", \
                            dtype_map_dict[requantize_node.attr['out_type'].type])
                        Helper.set_attr_dtype(new_node, "Tsummand", \
                            dtype_map_dict[requantize_node.attr['out_type'].type])
                    elif len(quantized_node.attr['fused_ops'].list.s) == 0:
                        Helper.set_attr_type_list(new_node, 'input_types', [
                            input_data_type.as_datatype_enum,
                            dtypes.qint8.as_datatype_enum,
                            #dtypes.float32.as_datatype_enum if new_node.attr["Tbias"].type == dtypes.float32 \
                                                                            # else dtypes.qint32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                        ])
                        self.fused_ops= [b"Requantize"]
                        Helper.set_attr_type_list(new_node, 'out_types', [
                                              requantize_node.attr['out_type'].type,
                                              dtypes.float32.as_datatype_enum,
                                              dtypes.float32.as_datatype_enum ])
                        Helper.set_attr_dtype(new_node, "out_type", \
                            dtype_map_dict[requantize_node.attr['out_type'].type])
                        Helper.set_attr_dtype(new_node, "Tsummand", \
                            dtype_map_dict[requantize_node.attr['out_type'].type])
                Helper.set_attr_string_list(new_node, 'fused_ops', self.fused_ops)
     
            if "_kernel" in quantized_node.attr:
                new_node.attr["_kernel"].CopyFrom(quantized_node.attr['_kernel'])

            if new_node.op in ('_QuantizedConv3D'):
                input_data_type = dtypes.qint8 if new_node.attr["Tinput"].type == dtypes.qint8 else dtypes.quint8
                if len(quantized_node.attr['fused_ops'].list.s) == 0:
                    Helper.set_attr_string_list(new_node, 'fused_ops', [ b'Requantize'])
                    Helper.set_attr_type_list(new_node, 'input_types', [
                        input_data_type.as_datatype_enum,
                        dtypes.qint8.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ])
                    Helper.set_attr_type_list(new_node, 'out_types', [
                                          requantize_node.attr['out_type'].type,
                                          dtypes.float32.as_datatype_enum,
                                          dtypes.float32.as_datatype_enum ])
                    Helper.set_attr_dtype(new_node, "out_type", \
                        dtype_map_dict[requantize_node.attr['out_type'].type])
                    Helper.set_attr_dtype(new_node, "Tsummand", \
                        dtype_map_dict[requantize_node.attr['out_type'].type])
                elif str(quantized_node.attr['fused_ops'].list.s) == str([b"BiasAdd"]):
                    Helper.set_attr_string_list(new_node, 'fused_ops', [b'BiasAdd', b'Requantize'])
                    Helper.set_attr_type_list(new_node, 'input_types', [
                        input_data_type.as_datatype_enum,
                        dtypes.qint8.as_datatype_enum,
                        dtypes.float32.as_datatype_enum if new_node.attr["Tbias"].type == dtypes.float32 else \
                                                                                    dtypes.qint32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ])
                    Helper.set_attr_type_list(new_node, 'out_types', [
                                          requantize_node.attr['out_type'].type,
                                          dtypes.float32.as_datatype_enum,
                                          dtypes.float32.as_datatype_enum ])
                    Helper.set_attr_dtype(new_node, "out_type", \
                        dtype_map_dict[requantize_node.attr['out_type'].type])
                    Helper.set_attr_dtype(new_node, "Tsummand", \
                        dtype_map_dict[requantize_node.attr['out_type'].type])
                elif str(quantized_node.attr['fused_ops'].list.s) == str([b"BiasAdd", b"Relu"]):
                    Helper.set_attr_string_list(new_node, 'fused_ops', [b'BiasAdd', b'Relu', b'Requantize'])
                    Helper.set_attr_type_list(new_node, 'input_types', [
                        input_data_type.as_datatype_enum,
                        dtypes.qint8.as_datatype_enum,
                        dtypes.float32.as_datatype_enum if new_node.attr["Tbias"].type == dtypes.float32 else \
                                                                                    dtypes.qint32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ])
                    Helper.set_attr_type_list(new_node, 'out_types', [
                                          requantize_node.attr['out_type'].type,
                                          dtypes.float32.as_datatype_enum,
                                          dtypes.float32.as_datatype_enum ])
                    Helper.set_attr_dtype(new_node, "out_type", \
                        dtype_map_dict[requantize_node.attr['out_type'].type])
                    Helper.set_attr_dtype(new_node, "Tsummand", \
                        dtype_map_dict[requantize_node.attr['out_type'].type])
                 
            if quantized_node.op == "QuantizedConv2D" or \
                    quantized_node.op == "QuantizedConv2DWithBias" or \
                    quantized_node.op == "QuantizedDepthwiseConv2D" or \
                    quantized_node.op == "QuantizedDepthwiseConv2DWithBias" or \
                    ('alpha' in quantized_node.attr and quantized_node.attr['alpha'].f > 0):
                new_node.attr["out_type"].CopyFrom(attr_value_pb2.AttrValue(type=int8_type))
            elif quantized_node.op == "QuantizedConv2DWithBiasAndRelu" or \
                 quantized_node.op == "QuantizedDepthwiseConv2DWithBiasAndRelu":
                new_node.attr["out_type"].CopyFrom(attr_value_pb2.AttrValue(type=uint8_type))
            elif new_node.op not in ('_QuantizedConv2D', '_QuantizedDepthwiseConv2D', '_QuantizedConv3D'):
                new_node.attr["out_type"].CopyFrom(attr_value_pb2.AttrValue(type=uint8_type))

            old_input_name = dequantize_node_name if dequantize_node_name else requantize_node_name
            self.graph_analyzer.replace_single_node(
                new_node, [parent_node_name], quantized_node_name,
                self.graph_info[old_input_name].outputs, old_input_name)
            self.graph_analyzer.remove_node(quantized_node_name)

        target_nodes = self.graph_analyzer.query_fusion_pattern_nodes(self.sum_pattern)
        for i in target_nodes:
            quantized_node_name = i[0]
            quantized_node = self.graph_info[quantized_node_name].node
            if i[-1][0] in ('_QuantizedDepthwiseConv2D', '_QuantizedConv2D', '_QuantizedConv3D'):
                if quantized_node.attr['fused_ops'].list.s not in self.fuse_sum_op_types:
                   continue
                #else:
                #   print(quantized_node.attr['fused_ops'].list.s)

            requantize_node_name = i[1]
            requantize_node = self.graph_info[requantize_node_name].node
            requested_output_min_name = requantize_node.input[3]
            requested_output_max_name = requantize_node.input[4]

            quantized_node_op = i[-1][0]

            new_node = node_def_pb2.NodeDef()

            if self.new_api:
                if i[-1][0] in ('QuantizedConv2DWithBiasSumAndRelu',) :
                    new_node.op = '_QuantizedConv2D'
                    self.fused_ops= [b"BiasAdd", b"Sum", b"Relu", b"Requantize"]
                elif i[-1][0] == '_QuantizedConv2D':
                    new_node.op = '_QuantizedConv2D'
                elif quantized_node_op == '_QuantizedDepthwiseConv2D':
                    new_node.op = '_QuantizedDepthwiseConv2D'
                elif i[-1][0] == '_QuantizedConv3D':
                    new_node.op = '_QuantizedConv3D'
            else:
                new_node.op = quantized_node_op + "AndRequantize"

            new_node.name = requantize_node_name

            for _, value in enumerate(quantized_node.input[:-1]):
                new_node.input.append(value)

            new_node.attr["Tinput"].CopyFrom(quantized_node.attr['Tinput'])
            new_node.attr["Tfilter"].CopyFrom(quantized_node.attr['Tfilter'])
            new_node.attr["strides"].CopyFrom(quantized_node.attr['strides'])
            new_node.attr["padding"].CopyFrom(quantized_node.attr['padding'])
            #new_node.attr["Tsummand"].CopyFrom(quantized_node.attr['Tsummand'])

            new_node.input.append(requested_output_min_name)
            new_node.input.append(requested_output_max_name)
            deq_node = self.graph_info[Helper.node_name_from_input(quantized_node.input[-1])].node
            if deq_node.op != 'Dequantize' or deq_node.op.find("Quantize") != -1:
                continue
      
            if deq_node.op == 'Dequantize':
                original_summand_node = self.graph_info[Helper.node_name_from_input(
                    deq_node.input[0])].node
            else:
                original_summand_node = deq_node
            summand_op_type = uint8_type if dtypes.as_dtype(
                deq_node.attr["T"].type) == uint8_type else int8_type

            for j in range(3):
                new_node.input.append(original_summand_node.name + ':{}'.format(j))
            # in tf 2.10, the "padding_list" attr name changes to explicit_paddings
            if "padding_list" in quantized_node.attr:
                if not self.new_api:
                    new_node.attr["padding_list"].CopyFrom(quantized_node.attr['padding_list'])
                elif quantized_node.attr["padding"].s == b"EXPLICIT":
                    new_node.attr["explicit_paddings"].CopyFrom(quantized_node.attr['padding_list'])
            elif "explicit_paddings" in quantized_node.attr:
                new_node.attr["explicit_paddings"].CopyFrom(quantized_node.attr['explicit_paddings'])

            if "dilations" in quantized_node.attr:
                new_node.attr["dilations"].CopyFrom(quantized_node.attr['dilations'])

            if "alpha" in quantized_node.attr and quantized_node.attr['alpha'].f > 0:
                new_node.attr["out_type"].CopyFrom(attr_value_pb2.AttrValue(type=int8_type))
            else:
                new_node.attr["out_type"].CopyFrom(attr_value_pb2.AttrValue(type=uint8_type))

            new_node.attr["Tbias"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type))
            new_node.attr["Tsummand"].CopyFrom(attr_value_pb2.AttrValue(type=summand_op_type))
 
            if new_node.op in ('_QuantizedConv2D', '_QuantizedDepthwiseConv2D', '_QuantizedConv3D'):
                original_input = list(new_node.input)
                new_input = []
                new_input.extend(original_input[:3])
                new_input.append(original_input[-3])
                new_input.extend(original_input[3:7])
                new_input.extend(original_input[-2:])
                new_input.extend(original_input[-5:-3])
                new_node.ClearField('input')
                new_node.input.extend(new_input)
                input_data_type = dtypes.qint8 if new_node.attr["Tinput"].type == dtypes.qint8 else dtypes.quint8
                Helper.set_attr_type_list(new_node, 'input_types', [
                    input_data_type.as_datatype_enum,
                    dtypes.qint8.as_datatype_enum,
                    dtypes.float32.as_datatype_enum,
                    dtypes.quint8.as_datatype_enum if summand_op_type != int8_type else dtypes.qint8.as_datatype_enum,
                    dtypes.float32.as_datatype_enum,
                    dtypes.float32.as_datatype_enum,
                    dtypes.float32.as_datatype_enum,
                    dtypes.float32.as_datatype_enum,
                    dtypes.float32.as_datatype_enum,
                    dtypes.float32.as_datatype_enum,
                    dtypes.float32.as_datatype_enum,
                    dtypes.float32.as_datatype_enum
                ])
                Helper.set_attr_dtype(new_node, "Tsummand", dtypes.quint8 if summand_op_type != int8_type \
                   else dtypes.qint8)
                if str(quantized_node.attr['fused_ops'].list.s) == str([b'BiasAdd', b'Sum', b'Relu']):
                    self.fused_ops = [b'BiasAdd', b'Sum', b'Relu', b'Requantize']
                    Helper.set_attr_type_list(new_node, 'out_types', [
                                          requantize_node.attr['out_type'].type,
                                          dtypes.float32.as_datatype_enum,
                                          dtypes.float32.as_datatype_enum ])
                    Helper.set_attr_dtype(new_node, "out_type", \
                        dtype_map_dict[requantize_node.attr['out_type'].type])
                elif str(quantized_node.attr['fused_ops'].list.s) == str([b'BiasAdd', b'Relu', b'Sum']):
                    self.fused_ops = [b'BiasAdd', b'Relu', b'Sum', b'Requantize']
                    Helper.set_attr_type_list(new_node, 'out_types', [
                                          requantize_node.attr['out_type'].type,
                                          dtypes.float32.as_datatype_enum,
                                          dtypes.float32.as_datatype_enum ])
                    Helper.set_attr_dtype(new_node, "out_type", \
                        dtype_map_dict[requantize_node.attr['out_type'].type])
                elif str(quantized_node.attr['fused_ops'].list.s) == str([b'BiasAdd', b'LeakyRelu', b'Sum']):
                    self.fused_ops = [b'BiasAdd', b'LeakyRelu', b'Sum', b'Requantize']
                    Helper.set_attr_type_list(new_node, 'out_types', [
                                          requantize_node.attr['out_type'].type,
                                          dtypes.float32.as_datatype_enum,
                                          dtypes.float32.as_datatype_enum ])
                    Helper.set_attr_dtype(new_node, "out_type", \
                        dtype_map_dict[requantize_node.attr['out_type'].type])
                    #Current fusion requires summand has same dtype as output if output is qint8
                    Helper.set_attr_dtype(new_node, "Tsummand", \
                        dtype_map_dict[requantize_node.attr['out_type'].type])

                elif str(quantized_node.attr['fused_ops'].list.s) == str([b'BiasAdd', b'Sum']):
                    self.fused_ops = [b'BiasAdd', b'Sum', b'Requantize']
                    Helper.set_attr_type_list(new_node, 'out_types', [
                                          requantize_node.attr['out_type'].type,
                                          dtypes.float32.as_datatype_enum,
                                          dtypes.float32.as_datatype_enum ])
                    Helper.set_attr_dtype(new_node, "out_type", \
                        dtype_map_dict[requantize_node.attr['out_type'].type])
                    #Current fusion requires summand has same dtype as output if output is qint8
                    Helper.set_attr_dtype(new_node, "Tsummand", \
                        dtype_map_dict[requantize_node.attr['out_type'].type])
                Helper.set_attr_string_list(new_node, 'fused_ops', self.fused_ops)

            if not self.new_api:
                if quantized_node_op == 'QuantizedConv2DWithBiasReluAndSum':
                    new_node.op = 'QuantizedConv2DWithBiasReluAndSumAndRequantize'
                    if "alpha" in quantized_node.attr:
                        new_node.attr["alpha"].CopyFrom(quantized_node.attr['alpha'])

                elif summand_op_type == int8_type:
                    new_node.op = "QuantizedConv2DWithBiasSignedSumAndReluAndRequantize"

            self.graph_analyzer.replace_single_node(
                new_node, [quantized_node.input[0], original_summand_node.name],
                quantized_node.name, self.graph_info[requantize_node_name].outputs,
                requantize_node_name)
            self.graph_analyzer.remove_node(quantized_node_name)

            if deq_node.op == 'Dequantize':
                self.graph_analyzer.remove_node_with_single_input_output(deq_node.name)

        return self.graph_analyzer.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import csv
import time
import uuid
from datetime import datetime
from typing import Tuple

import numpy as np
import ofa
from fvcore.nn import FlopCountAnalysis
from ofa.imagenet_classification.data_providers.imagenet import ImagenetDataProvider
from ofa.imagenet_classification.run_manager import ImagenetRunConfig, RunManager
from ofa.tutorial.flops_table import rm_bn_from_net

from neural_compressor.experimental.nas.dynast.dynas_manager import ParameterManager
from neural_compressor.experimental.nas.dynast.dynas_predictor import Predictor
from neural_compressor.utils.utility import LazyImport, logger

torch = LazyImport('torch')
torchvision = LazyImport('torchvision')


def get_macs(
    model: torch.nn.Module,
    input_size: Tuple[int, int, int, int] = (1, 3, 224, 224),
    device: str = 'cpu',
) -> int:
    model = model.to(device)
    rm_bn_from_net(model)
    model.eval()
    inputs = torch.randn(*input_size, device=device)
    macs = FlopCountAnalysis(model, inputs).total() // input_size[0]
    return macs


def _auto_steps(
    batch_size: int,
    is_warmup: bool = False,
    warmup_scale: float = 5.0,
    min_steps: int = 25,
    min_samples: int = 500,
) -> int:
    """Simple scaling of number of steps w.r.t batch_size

    Example:
    1. `_auto_steps(1, True), _auto_steps(1, False)` -> 100, 500
    2. `_auto_steps(8, True), _auto_steps(8, False)` -> 12, 62
    3. `_auto_steps(16, True), _auto_steps(8, False)` -> 6, 31
    4. `_auto_steps(32, True), _auto_steps(8, False)` -> 5, 25

    Args:
    - batch_size
    - is_warmup: if set to True, will scale down the number of steps by `warmup_scale`.
    - warmup_scale: scale by which number of steps should be decreased if `is_warmup` is True.
    - min_steps: minimum number of steps to return
    - min_samples: returned steps multiplied by `batch_size` should be at least this much.

    Returns:
        number of steps
    """

    if not is_warmup:
        warmup_scale = 1.0

    return int(max(batch_size*min_steps, min_samples)//batch_size//warmup_scale)


@torch.no_grad()
def measure_latency(
    model: torch.nn.Module,
    input_size: Tuple[int, int, int, int] = (1, 3, 224, 224),
    warmup_steps: int = None,
    measure_steps: int = None,
    device: str = 'cpu',
) -> Tuple[float, float]:
    """Measure Torch model's latency.

    Args:
    - model: Torch model
    - input_size: a tuple (batch size, channels, resolution, resolution).
    - warmup_steps - how many data batches to use to warm up the device.
        If 'None' it will be adjusted automatically w.r.t batch size.
    - measure_steps - how many data batches to use for latency measurement.
        If 'None' it will be adjusted automatically w.r.t batch size.
    - device: which device is being used for latency measurement.

    Returns:
        mean latency; std latency
    """
    if not warmup_steps:
        warmup_steps = _auto_steps(input_size[0], is_warmup=True)
    if not measure_steps:
        measure_steps = _auto_steps(input_size[0])

    times = []

    inputs = torch.randn(input_size, device=device)
    model = model.eval()
    rm_bn_from_net(model)
    model = model.to(device)

    if 'cuda' in str(device):
        torch.cuda.synchronize()
    for _ in range(warmup_steps):
        model(inputs)
    if 'cuda' in str(device):
        torch.cuda.synchronize()

    for _ in range(measure_steps):
        if 'cuda' in str(device):
            torch.cuda.synchronize()
        st = time.time()
        model(inputs)
        if 'cuda' in str(device):
            torch.cuda.synchronize()
        ed = time.time()
        times.append(ed - st)

    # Convert to s->ms, round to 0.001
    latency_mean = np.round(np.mean(times) * 1e3, 3)
    latency_std = np.round(np.std(times) * 1e3, 3)

    return latency_mean, latency_std


class Runner:
    pass


class OFARunner(Runner):
    """The OFARunner class manages the sub-network selection from the OFA super-network and
    the validation measurements of the sub-networks. ResNet50, MobileNetV3 w1.0, and MobileNetV3 w1.2
    are currently supported. Imagenet is required for these super-networks `imagenet-ilsvrc2012`.
    """

    def __init__(
        self,
        supernet: str,
        acc_predictor: Predictor,
        macs_predictor: Predictor,
        latency_predictor: Predictor,
        imagenetpath: str,
        batch_size: int,
    ) -> None:
        self.supernet = supernet
        self.acc_predictor = acc_predictor
        self.macs_predictor = macs_predictor
        self.latency_predictor = latency_predictor
        self.device = 'cpu'
        self.test_size = None
        ImagenetDataProvider.DEFAULT_PATH = imagenetpath
        self.ofa_network = ofa.model_zoo.ofa_net(supernet, pretrained=True)
        self.run_config = ImagenetRunConfig(test_batch_size=64, n_worker=20)
        self.batch_size = batch_size

    def estimate_accuracy_top1(
        self,
        subnet_cfg: dict,
    ) -> float:
        top1 = self.acc_predictor.predict(subnet_cfg)
        return top1

    def estimate_macs(
        self,
        subnet_cfg: dict,
    ) -> int:
        macs = self.macs_predictor.predict(subnet_cfg)
        return macs

    def estimate_latency(
        self,
        subnet_cfg: dict,
    ) -> float:
        latency = self.latency_predictor.predict(subnet_cfg)
        return latency

    def validate_top1(
        self,
        subnet_cfg: dict,
    ) -> float: # pragma: no cover
        subnet = self.get_subnet(subnet_cfg)
        folder_name = '.torch/tmp-{}'.format(uuid.uuid1().hex)
        run_manager = RunManager(
            '{}/eval_subnet'.format(folder_name), subnet, self.run_config, init=False
        )
        run_manager.reset_running_statistics(net=subnet)

        # Test sampled subnet
        self.run_config.data_provider.assign_active_img_size(subnet_cfg['r'][0])
        loss, acc = run_manager.validate(net=subnet, no_logs=False)
        top1 = acc[0]
        return top1

    def validate_macs(
        self,
        subnet_cfg: dict,
    ) -> float:
        """Measure Torch model's FLOPs/MACs as per FVCore calculation
        Args:
            subnet_cfg: sub-network Torch model
        Returns:
            `macs`
        """

        model = self.get_subnet(subnet_cfg)
        input_size = (self.batch_size, 3, 224, 224)
        macs = get_macs(model=model, input_size=input_size, device=self.device)
        logger.info('Model\'s macs: {}'.format(macs))
        return macs

    @torch.no_grad()
    def measure_latency(
        self,
        subnet_cfg: dict,
        warmup_steps: int = None,
        measure_steps: int = None,
    ) -> Tuple[float, float]:
        """Measure OFA model's latency.
        Args:
            subnet_cfg: sub-network Torch model
        Returns:
            mean latency; std latency
        """
        model = self.get_subnet(subnet_cfg)
        input_size = (self.batch_size, 3, 224, 224)

        latency_mean, latency_std = measure_latency(
            model=model,
            input_size=input_size,
            warmup_steps=warmup_steps,
            measure_steps=measure_steps,
            device=self.device,
        )
        logger.info('Model\'s latency: {} +/- {}'.format(latency_mean, latency_std))

        return latency_mean, latency_std

    def get_subnet(
        self,
        subnet_cfg: dict,
    ) -> torch.nn.Module:
        if self.supernet == 'ofa_resnet50':
            self.ofa_network.set_active_subnet(
                ks=subnet_cfg['d'], e=subnet_cfg['e'], d=subnet_cfg['w']
            )
        else:
            self.ofa_network.set_active_subnet(
                ks=subnet_cfg['ks'], e=subnet_cfg['e'], d=subnet_cfg['d']
            )

        self.subnet = self.ofa_network.get_active_subnet(preserve_weight=True)
        self.subnet.eval()
        return self.subnet


class EvaluationInterface:
    """
    The interface class update is required to be updated for each unique SuperNetwork
    framework as it controls how evaluation calls are made from DyNAS-T

    Args:
        evaluator : class
            The 'runner' that performs the validation or prediction
        manager : class
            The DyNAS-T manager that translates between PyMoo and the parameter dict
        csv_path : string
            (Optional) The csv file that get written to during the subnetwork search
    """

    def __init__(
        self,
        evaluator: Runner,
        manager: ParameterManager,
        metrics: list = ['acc', 'macs'],
        predictor_mode: bool = False,
        csv_path: str = None,
    ) -> None:
        self.evaluator = evaluator
        self.manager = manager
        self.metrics = metrics
        self.predictor_mode = predictor_mode
        self.csv_path = csv_path

    def eval_subnet(
        self,
        x: list,
    ) -> Tuple[dict, float, float]:
        pass

    def clear_csv(self) -> None:
        if self.csv_path:
            f = open(self.csv_path, "w")
            writer = csv.writer(f)
            result = ['Sub-network', 'Date', 'Latency (ms)', ' MACs', 'Top-1 Acc (%)']
            writer.writerow(result)
            f.close()


class EvaluationInterfaceResNet50(EvaluationInterface):
    def __init__(
        self,
        evaluator: Runner,
        manager: ParameterManager,
        metrics: list = ['acc', 'macs'],
        predictor_mode: bool = False,
        csv_path: str = None,
    ) -> None:
        super().__init__(evaluator, manager, metrics, predictor_mode, csv_path)

    def eval_subnet(
        self,
        x: list,
    ) -> Tuple[dict, float, float]:
        # PyMoo vector to Elastic Parameter Mapping
        param_dict = self.manager.translate2param(x)

        sample = {
            'd': param_dict['d'],
            'e': param_dict['e'],
            'w': param_dict['w'],
            'r': [224],
        }
        subnet_sample = copy.deepcopy(sample)

        # Always evaluate/predict top1
        lat, macs = 0, 0
        if self.predictor_mode == True:
            top1 = self.evaluator.estimate_accuracy_top1(
                self.manager.onehot_generic(x).reshape(1, -1)
            )[0]
            if 'macs' in self.metrics:
                macs = self.evaluator.estimate_macs(
                    self.manager.onehot_generic(x).reshape(1, -1)
                )[0]
            if 'lat' in self.metrics:
                lat = self.evaluator.estimate_latency(
                    self.manager.onehot_generic(x).reshape(1, -1)
                )[0]
        else:
            top1 = self.evaluator.validate_top1(subnet_sample)
            macs = self.evaluator.validate_macs(subnet_sample)
            if 'lat' in self.metrics:
                lat, _ = self.evaluator.measure_latency(subnet_sample)

        if self.csv_path:
            with open(self.csv_path, 'a') as f:
                writer = csv.writer(f)
                date = str(datetime.now())
                result = [subnet_sample, date, lat, macs, top1]
                writer.writerow(result)

        # PyMoo only minimizes objectives, thus accuracy needs to be negative
        # Requires format: subnetwork, objective x, objective y
        if 'lat' in self.metrics:
            return sample, lat, -top1
        else:
            return sample, macs, -top1


class EvaluationInterfaceMobileNetV3(EvaluationInterface):
    def __init__(
        self,
        evaluator: Runner,
        manager: ParameterManager,
        metrics=['acc', 'macs'],
        predictor_mode=False,
        csv_path=None,
    ) -> None:
        super().__init__(evaluator, manager, metrics, predictor_mode, csv_path)

    def eval_subnet(
        self,
        x: list,
    ) -> Tuple[dict, float, float]:
        # PyMoo vector to Elastic Parameter Mapping
        param_dict = self.manager.translate2param(x)

        sample = {
            'wid': None,
            'ks': param_dict['ks'],
            'e': param_dict['e'],
            'd': param_dict['d'],
            'r': [224],
        }
        subnet_sample = copy.deepcopy(sample)

        # Always evaluate/predict top1
        lat, macs = 0, 0
        if self.predictor_mode == True:
            top1 = self.evaluator.estimate_accuracy_top1(self.manager.onehot_generic(x).reshape(1,-1))[0]
            if 'macs' in self.metrics:
                macs = self.evaluator.estimate_macs(self.manager.onehot_generic(x).reshape(1,-1))[0]
            if 'lat' in self.metrics:
                lat = self.evaluator.estimate_latency(self.manager.onehot_generic(x).reshape(1,-1))[0]
        else:
            top1 = self.evaluator.validate_top1(subnet_sample)
            macs = self.evaluator.validate_macs(subnet_sample)
            if 'lat' in self.metrics:
                lat, _ = self.evaluator.measure_latency(subnet_sample)

        if self.csv_path:
            with open(self.csv_path, 'a') as f:
                writer = csv.writer(f)
                date = str(datetime.now())
                result = [subnet_sample, date, lat, macs, top1]
                writer.writerow(result)

        # PyMoo only minimizes objectives, thus accuracy needs to be negative
        # Requires format: subnetwork, objective x, objective y
        if 'lat' in self.metrics:
            return sample, lat, -top1
        else:
            return sample, macs, -top1


def get_torchvision_model(
    model_name: str,
) -> torch.nn.Module:
    try:
        model = getattr(torchvision.models, model_name)(pretrained=True)
        model.eval()
        return model
    except AttributeError as ae: # pragma: no cover
        logger.error(
            'Model {model_name} not available. This can be due to either a typo or the model is not '
            'available in torchvision=={torchvision_version}. \nAvailable models: {available_models}'.format(
                model_name=model_name,
                torchvision_version=torchvision.__version__,
                available_models=', '.join(
                    [m for m in dir(torchvision.models) if not m.startswith('_')]
                ),
            )
        )
        raise ae


class TorchVisionReference:
    def __init__(
        self,
        model_name: str,
        dataset_path: str,
        batch_size: int,
        input_size: int = 224,
    ) -> None:
        if 'ofa_resnet50' in model_name:
            model_name = 'resnet50'
        if 'ofa_mbv3' in model_name:
            model_name = 'mobilenet_v3_large'

        self.model_name = model_name
        self.dataset_path = dataset_path
        self.batch_size = batch_size
        self.input_size = input_size

        logger.info(
            '{name} for \'{model_name}\' on \'{val_dataset_path}\' dataset'.format(
                name=str(self),
                model_name=self.model_name,
                val_dataset_path=self.dataset_path,
            )
        )

        # Here we only download the model. Model will have to be loaded for each validation and benchmarking call
        # separately to avoid modifications to the model being passed between calls.
        get_torchvision_model(model_name=self.model_name)

    def validate_top1(self) -> Tuple[float, float, float]: # pragma: no cover
        ImagenetDataProvider.DEFAULT_PATH = self.dataset_path
        model = get_torchvision_model(model_name=self.model_name)
        run_config = ImagenetRunConfig(test_batch_size=64, n_worker=20)
        folder_name = '.torch/tmp-{}'.format(uuid.uuid1().hex)
        run_manager = RunManager(
            '{}/eval_subnet'.format(folder_name), model, run_config, init=False
        )
        run_config.data_provider.assign_active_img_size(224)
        loss, acc = run_manager.validate(net=model, no_logs=False)
        top1, top5 = acc
        return loss, top1, top5

    def validate_macs(
        self,
        device: str = 'cpu',
    ) -> int:
        model = get_torchvision_model(model_name=self.model_name)
        input_size = (self.batch_size, 3, self.input_size, self.input_size)

        macs = get_macs(model=model, input_size=input_size, device=device)
        logger.info(
            '\'{model_name}\' macs {macs}'.format(
                model_name=self.model_name,
                macs=macs,
            )
        )
        return macs

    def measure_latency(
        self,
        device: str = 'cpu',
        warmup_steps: int = None,
        measure_steps: int = None,
    ) -> Tuple[float, float]:
        model = get_torchvision_model(model_name=self.model_name)
        input_size = (self.batch_size, 3, self.input_size, self.input_size)

        latency_mean, latency_std = measure_latency(
            model=model,
            input_size=input_size,
            warmup_steps=warmup_steps,
            measure_steps=measure_steps,
            device=device,
        )
        logger.info(
            '\'{model_name}\' mean latency {latency_mean} +/- {latency_std}'.format(
                model_name=self.model_name,
                latency_mean=latency_mean,
                latency_std=latency_std,
            )
        )
        return latency_mean, latency_std



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from neural_compressor.utils.utility import singleton

@singleton
class TensorflowFilters(object):
    def __init__(self):
        self.filters = {}
        self.filters.update(TENSORFLOW_FILTERS)

@singleton
class ONNXRTQLFilters(object):
    def __init__(self):
        self.filters = {}
        self.filters.update(ONNXRT_QL_FILTERS)

@singleton
class ONNXRTITFilters(object):
    def __init__(self):
        self.filters = {}
        self.filters.update(ONNXRT_IT_FILTERS)

@singleton
class PyTorchFilters(object):
    def __init__(self):
        self.filters = {}
        self.filters.update(PYTORCH_FILTERS)

@singleton
class MXNetFilters(object):
    def __init__(self):
        self.filters = {}
        self.filters.update(MXNET_FILTERS)

TENSORFLOW_FILTERS = {}
INTELTENSORFLOW_FILTERS = {}
TENSORFLOW_ITEX_FILTERS = {}
ONNXRT_IT_FILTERS = {}
ONNXRT_QL_FILTERS = {}
PYTORCH_FILTERS = {}
MXNET_FILTERS = {}

framework_filters = {"tensorflow": TensorflowFilters,
                     "inteltensorflow": TensorflowFilters,
                     "tensorflow_itex": TensorflowFilters,
                     "pytorch": PyTorchFilters,
                     "pytorch_ipex": PyTorchFilters,
                     "pytorch_fx": PyTorchFilters,
                     "mxnet": MXNetFilters,
                     "onnxrt_qlinearops": ONNXRTQLFilters,
                     "onnxrt_qdq": ONNXRTQLFilters,
                     "onnxrt_qoperator": ONNXRTQLFilters,
                     "onnxrt_integerops": ONNXRTITFilters,
                     }

registry_filters = {"tensorflow": TENSORFLOW_FILTERS,
                    "inteltensorflow": INTELTENSORFLOW_FILTERS,
                    "tensorflow_itex": TENSORFLOW_ITEX_FILTERS,
                    "pytorch": PYTORCH_FILTERS,
                    "pytorch_ipex": PYTORCH_FILTERS,
                    "pytorch_fx": PYTORCH_FILTERS,
                    "mxnet": MXNET_FILTERS,
                    "onnxrt_integerops": ONNXRT_IT_FILTERS,
                    "onnxrt_qdq": ONNXRT_QL_FILTERS,
                    "onnxrt_qoperator": ONNXRT_QL_FILTERS,
                    "onnxrt_qlinearops": ONNXRT_QL_FILTERS}

class FILTERS(object):
    def __init__(self, framework):
        assert framework in ["tensorflow", "inteltensorflow", "tensorflow_itex",
                             "mxnet", "onnxrt_qdq", "pytorch", "pytorch_ipex", "pytorch_fx", 
                             "onnxrt_integerops", "onnxrt_qlinearops", "onnxrt_qoperator"], \
                             "framework support tensorflow pytorch mxnet onnxrt"
        self.filters = framework_filters[framework]().filters
        self.framework = framework

    def __getitem__(self, filter_type):
        assert filter_type in self.filters.keys(), "filter support {}".\
            format(self.filters.keys())
        return self.filters[filter_type]


def filter_registry(filter_type, framework):
    """The class decorator used to register all transform subclasses.


    Args:
        filter_type (str): fILTER registration name
        framework (str): support 4 framework including 'tensorflow', 'pytorch', 'mxnet', 'onnxrt'
        cls (class): The class of register.

    Returns:
        cls: The class of register.
    """
    def decorator_transform(cls):
        for single_framework in [fwk.strip() for fwk in framework.split(',')]:
            assert single_framework in [
                "tensorflow", 
                "inteltensorflow",
                "tensorflow_itex",
                "pytorch",
                "pytorch_ipex",
                "pytorch_fx",
                "mxnet",
                "onnxrt_integerops",
                "onnxrt_qdq",
                "onnxrt_qlinearops",
                "onnxrt_qoperator"
            ], "The framework support tensorflow mxnet pytorch onnxrt"
            if filter_type in registry_filters[single_framework].keys():
                raise ValueError('Cannot have two transforms with the same name')
            registry_filters[single_framework][filter_type] = cls
        return cls
    return decorator_transform

class Filter(object):
    """The base class for transform. __call__ method is needed when write user specific transform

    """
    @abstractmethod
    def __call__(self, *args, **kwargs):
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..conf.dotdict import DotDict

class onnxrt:
    graph_optimization = DotDict({'level': None, 'gemm2matmul': True})
    qdq_setting = DotDict({'OpTypesToExcludeOutputQuantizatioin': None, 
                           'AddQDQPairToWeight': False,
                           'DedicatedQDQPair': False})

OPTIONS = {'tensorflow': None,
           'inteltensorflow': None,
           'tensorflow_itex': None,
           'pytorch': None,
           'pytorch_fx': None,
           'pytorch_ipex': None,
           'mxnet': None,
           'onnxrt_integerops': onnxrt,
           'onnxrt_qlinearops': onnxrt,
           'onnxrt_qdq': onnxrt,
           'onnxrt_qoperator': onnxrt,
           }





#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pickle
import random
import tempfile
import sys
import numpy as np
import yaml
from ..conf.config import MixedPrecision_Conf
from ..conf.dotdict import deep_get, deep_set, DotDict
from ..strategy import STRATEGIES
from ..utils import logger
from ..utils.create_obj_from_config import create_dataloader
from ..utils.utility import CpuInfo, time_limit, set_backend
from .common import Model as NCModel
from ..model import BaseModel
from .graph_optimization import GraphOptimization

class MixedPrecision(GraphOptimization):
    """MixedPrecision class automatically generates low precision model across various DL
       frameworks including tensorflow, pytorch and onnxruntime.

    Args:
        conf_fname_or_obj (string or obj): The path to the YAML configuration file or
            Graph_Optimization_Conf class containing accuracy goal, tuning objective and
            preferred calibration & quantization tuning space etc.

    """

    def __init__(self, conf_fname_or_obj=None):
        self.conf_name = conf_fname_or_obj
        self._model = None
        self._eval_dataloader = None
        self._eval_func = None
        self._precisions = 'fp32'
        self._input = None
        self._output = None
        self.conf = None

        if isinstance(conf_fname_or_obj, MixedPrecision_Conf):
            self.conf = conf_fname_or_obj
        else:
            self.conf = MixedPrecision_Conf(conf_fname_or_obj)
        cfg = self.conf.usr_cfg
        if cfg.model.framework != 'NA':
            self.framework = cfg.model.framework.lower()
            set_backend(self.framework)

        cfg.tuning.strategy.name = 'automixedprecision'
        seed = cfg.tuning.random_seed
        random.seed(seed)
        np.random.seed(seed)

    def __call__(self):
        """The main entry point of mixed precision process.

           This interface works on all the DL frameworks that neural_compressor supports
           and provides three usages:
           a) Fully yaml configuration: User specifies all the info through yaml,
              including dataloaders used in calibration and evaluation phases
              and quantization tuning settings.

              For this usage, only model parameter is mandatory.

           b) Partial yaml configuration: User specifies dataloaders used in evaluation phase by code.
              The tool provides built-in dataloaders and evaluators, user just need provide
              a dataset implemented __iter__ or __getitem__ methods and invoke dataloader()
              with dataset as input parameter to create neural_compressor dataloader before calling this
              function.

              After that, User specifies fp32 "model" and evaluation dataset "eval_dataloader".
              The converted model is evaluated with "eval_dataloader" with evaluation metrics
              specified in the configuration file. The evaluation tells the tuner whether the
              converted model meets the accuracy criteria. If not, the tuner starts a tuning flow.

              For this usage, model and eval_dataloader parameters are mandatory.

        Returns:
            converted model: best converted model found, otherwise return None

        """

        assert isinstance(self._model, BaseModel), 'need set your Model for mixed precision....'
        if 'onnx' in self.framework and 'bf16' in self._precisions:
            logger.warning("Mixed precision doesn't support bf16 for ONNX models.")
            sys.exit(0)
        elif 'onnx' not in self.framework and 'fp16' in self._precisions:
            logger.warning("Mixed precision doesn't support fp16 for {} models.".format(
                self.framework))
            sys.exit(0)

        if self._precisions == ['bf16'] and not CpuInfo().bf16: # pragma: no cover
            if os.getenv('FORCE_BF16') == '1':
                logger.warning("Mixed precision will generate bf16 graph although " \
                               "the hardware doesn't support bf16 instruction.")
            else:
                logger.warning("Mixed precision exits due to the hardware " \
                               "doesn't support bf16 instruction.")
                sys.exit(0)

        if self._precisions == ['fp16'] and self.conf.usr_cfg.device != 'gpu':
            if os.getenv('FORCE_FP16') == '1':
                logger.warning("Mixed precision will generate fp16 graph although " \
                               "the hardware doesn't support fp16 instruction.")
            else:
                logger.warning("Mixed precision exits due to the hardware " \
                               "doesn't support fp16 instruction.")
                sys.exit(0)

        cfg = self.conf.usr_cfg
        if self.framework == 'tensorflow':
            self._model.name = cfg.model.name
            self._model.output_tensor_names = cfg.model.outputs if \
                not self._output else self._output
            self._model.input_tensor_names = cfg.model.inputs if \
                not self._input else self._input
            self._model.workspace_path = cfg.tuning.workspace.path

        # when eval_func is set, will be directly used and eval_dataloader can be None
        if self._eval_func is None:
            if self._eval_dataloader is None:
                eval_dataloader_cfg = deep_get(cfg, 'evaluation.accuracy.dataloader')
                if eval_dataloader_cfg is None:
                    self._eval_func = None
                else:
                    self._eval_dataloader = create_dataloader(self.framework, eval_dataloader_cfg)

        strategy = cfg.tuning.strategy.name.lower()

        assert strategy in STRATEGIES, "Tuning strategy {} is NOT supported".format(strategy)

        _resume = None
        # check if interrupted tuning procedure exists. if yes, it will resume the
        # whole auto tune process.
        self.resume_file = os.path.abspath(os.path.expanduser(cfg.tuning.workspace.resume)) \
                           if cfg.tuning.workspace and cfg.tuning.workspace.resume else None
        if self.resume_file: # pragma: no cover
            assert os.path.exists(self.resume_file), \
                "The specified resume file {} doesn't exist!".format(self.resume_file)
            with open(self.resume_file, 'rb') as f:
                _resume = pickle.load(f).__dict__

        self.strategy = STRATEGIES[strategy](
            self._model,
            self.conf,
            None,
            None,
            self._eval_dataloader,
            self._eval_func,
            _resume)

        try:
            with time_limit(self.conf.usr_cfg.tuning.exit_policy.timeout):
                self.strategy.traverse()
        except KeyboardInterrupt: # pragma: no cover
            pass
        except Exception as e: # pragma: no cover
            logger.info("Unexpected exception {} happened during turing.".format(repr(e)))
        finally: 
            if self.strategy.best_qmodel:
                logger.info(
                    "Specified timeout or max trials is reached! "
                    "Found a converted model which meet accuracy goal. Exit.")
                self.strategy.deploy_config()
            else: # pragma: no cover
                logger.info(
                    "Specified timeout or max trials is reached! "
                    "Not found any converted model which meet accuracy goal. Exit.")

            logger.info("Mixed Precision is done. Please invoke model.save() to save " \
                        "optimized model to disk.")

            return self.strategy.best_qmodel

    fit = __call__

    @property
    def precisions(self):
        return self._precisions

    @precisions.setter
    def precisions(self, customized_precisions):
        if isinstance(customized_precisions, list):
            self._precisions = sorted([i.strip() for i in customized_precisions])
        elif isinstance(customized_precisions, str):
            self._precisions = sorted([i.strip() for i in customized_precisions.split(',')])
        self.conf.usr_cfg.mixed_precision.precisions = self._precisions

    def set_config_by_model(self, model_obj):
        self.conf.usr_cfg.model.framework = model_obj.framework()
        if self._input:
            self.conf.usr_cfg.model.inputs = self._input
        if self._output:
            if isinstance(self._output, str) and ',' in self._output:
                self.conf.usr_cfg.model.outputs = [s.strip() for s in self._output.split(',')]
            else:
                self.conf.usr_cfg.model.outputs = self._output

    def __repr__(self):
        return 'MixedPrecision'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neural_compressor.utils.utility import LazyImport
from .base_dataloader import BaseDataLoader
import logging
mx = LazyImport('mxnet')

class MXNetDataLoader(BaseDataLoader):
    def _generate_dataloader(self, dataset, batch_size, last_batch, collate_fn,
                             sampler, batch_sampler, num_workers, pin_memory,
                             shuffle, distributed):
        if shuffle:
            logging.warning('Shuffle is not supported yet in MXNetDataLoader, ' \
                            'ignoring shuffle keyword.')
        return mx.gluon.data.DataLoader(
                dataset,
                batch_size=batch_size,
                batchify_fn=collate_fn,
                last_batch=last_batch,
                num_workers=num_workers,
                pin_memory=pin_memory,
                sampler=sampler,
                batch_sampler=batch_sampler)
